/*
   BILLGPC.CPP: Performs a bootp transaction using MS/TCP for Windows 95 or
   NT.  

   Free software Copyright (c) 1996 Lew Perin

   The program runs differently on 95 and NT, but the same executable
   runs on both OSes.

   The simplest case occurs both under NT and (when billgPC's companion
   driver for Windows 95 BILLGPS.VXD hasn't been installed) under 95: the
   outgoing bootp packet is transmitted using MS Winsock and the server's
   response is also awaited on a Winsock socket.

   Unfortunately this doesn't work on all networks.  When the bootpd fails
   to support RFC 1542 broadcasting, its response won't make it back through
   Microsoft's IP stack if the bootpd and client (i.e. us) disagree on the
   client's IP address.  If the IP stack on the bootpd's host does "Martian
   address filtering", then the bootpd probably won't see the request packet
   unless the client already knows what network (and subnet) it's on.

   Because of these problems, BILLGPS.VXD, when installed, fills two roles:

   - it, rather than Winsock, is used to receive the bootp reply;

   - it traps the sending of the request packet via Winsock and changes the
   source address to 0.0.0.0, adjusting the IP and UDP checksums accordingly.

   This program is indebted to the bootp client code in Erick Engelke's WATTCP.
   The idea for the project, as well as part of the design, grew out of
   discussions with Rich Graves, to whom the author also owes a big debt for
   finding bugs.  For the use of the Windows GUI, Mark Russinovich and
   Bryce Cogswell's VCMON served as a starting point.
*/

/*
   Revision History:

   Version  Date      Reason
   -------  --------
   0.21      5/12/96  Fixed string handling in hostname.
                      Fixed registry updating for case where there was no
                      previous value for a name.
                      Started using our first RFC 1533 code: domain name.
                      Gave the user more indications of the transaction's
                      progress.
                      Changed all cerr's to cout's to make it easier for
                      bug reporters to gather the evidence.
   0.30      5/19/96  GUIfied, i.e. got rid of all those couts!
                      The payoff of GUIfaction was to make it possible to
                      reboot if changes have been made to the Registry
                      and the user wishes it.
   0.31      5/21/96  Now we try to see if the machine's set up for DHCP
                      before trying a bootp transaction.  If it is DHCP, we
                      ask the user's permission to update the Registry for
                      bootp and immediately reboot.
                      Since we can't deal with multiple TCP/IP transport
                      keys, at least we detect the problem now and quit.
   0.32      5/23/96  Whoops! If only IPAddress got fixed, we weren't
                      offering to restart.  Thanks to Steve Rhoads.
   0.33      6/21/96  Whoops! We had omitted the 802.3 MAC length entry
                      in the RFC1042Header struct in shared.h.
   0.34       7/5/96  Now we don't emit as many messages about unhandled
                      vendor option codes.
                      Now we don't keep our window staring at the user for
                      30 seconds after we've finished our work; instead we
                      write the contents of the window's listbox to a log.
                      Started checking billgps.vxd version.
   0.35       7/8/96  Removed a race condition between the bootp sender and
                      receiver that could deafen us to a speedy server.
                      Fixed MessageBox owner window problem resulting in
                      occlusion of message box.
   0.36       7/9/96  We now dump a packet returned by the VXD that doesn't
                      satisfy us.
   0.37      7/19/96  Centralized user messages and logging, making it
                      possible to run minimized and pop up the listbox if
                      the user needs to be alerted.
                      Now we have our own icon!
   0.38      7/25/96  Take advantage of the Ndi/Default subkey of a TCP/IP
                      NetTrans key so we can deal with otherwise duplicative
                      TCP/IP NetTrans keys.
                      Set the current directory just in case we were launched
                      from the Registry ...\Run key.
   0.39       8/4/96  Fix bug found by Mikael Willberg in which the sender
                      thread continued broadcasting while the user considered
                      what to do after no valid response had been received.
   0.40       8/9/96  Installed the "/f" command line option to force fixing
                      Registry discrepancies without asking the user.
   0.41      8/15/96  We now use a version of the VxD that can handle
                      netcard drivers that truncate inbound bootp packets.
   0.42      8/21/96  Thanks to Patrick Audin and Jeff Mulherin, we now
                      try to get the MAC address by a call to NetBIOS.
                      Mikael Willberg found that our old, now fallback,
                      technique of using a Registry key wouldn't work
                      if the netcard had been replaced.
   0.43       9/1/96  We now use the "full transport" VxD.
   0.44       9/6/96  Now, on a machine with Dial-up Networking, we can
                      figure out which instance of TCP goes over the LAN
                      and ignore the DUN instance(s.)
   0.45     10/20/96  Now we try harder to get the real MAC address.  We
                      don't assume that the NetBIOS default LANA is the one
                      we want.  In fact, we prefer to get the MAC address
                      from our own VxD.  We've also cleaned up the way
                      we get the VxD version.  Now we also have a command-line
                      option "/p" (promiscuous) to dump the first packet
                      that arrives which satisfies criteria looser than
                      our normal ones.  We also now have a "/d" option that
                      simply dumps the guts of the driver.  We now allow
                      broadcast response packets because some sloppy bootp
                      servers send them.  Because bootpd/router combinations
                      may get confused, the VxD now no longer checks the
                      UDP source port.  We now pay attention to bootp hardware
                      type instead of assuming it's Ethernet.  With a bootpd
                      conforming to RFC 1542, we might be able to run without
                      a VxD now.  Introduced the "/i" runtime argument to
                      force an IEEE htype in the bootp request.  We now allow
                      the server to send us a huge vendor area.  Introduced
                      the "/r" runtime argument to force reboot if the Registry
                      is changed by the program.
   0.46     11/25/96  Now we'll also run under Windows NT.  We now log the
                      server address.  Changes to the VxD include a correction
                      of open error handling suggested by Dan Lanciani.
   0.47      1/27/97  Added the "/b" option to request a broadcast reply on
                      the first request; in conjunction with "/p" this shows
                      whether servers and routers are RFC 1542-compliant.
                      Win95 VxD now correctly handles token ring packets
                      devoid of routing info.  Introduced the "/t" option
                      for specifying the reply timeout in seconds.  When "/p"
                      is specified, we now tell the adapter to go into
                      promiscuous mode and we dump as many packets as qualify
                      until we time out.  Now we support a "/e" option to
                      set htype to Ethernet even if this is a lie.  Removed
                      an erroneous trailing comma that sometimes appeared in
                      processing the hostname.  When the new "/n" option is
                      used, we now update the Registry's ComputerName with
                      the hostname.  We now identify the version of the
                      protocol VxD if we use it.  With billgPS 1.21 we now
                      set the IP source address to 0.0.0.0 in our request
                      packet.
   0.48      3/29/97  To avoid hanging Win95/OSR2 we now use VxD level 1.22.
                      Because NT may start up TCP/IP raggedly, we retry bind()
                      multiple times between waits before giving up as long as
                      the error code encourages us.  Fixed bug affecting NT
                      machines with multiple name servers (thanks to Ian
                      Broadbent.)  Due to NT, became more careful in asking
                      for access as we navigate the Registry.  In 95 we can
                      now use the dynamic VxD of level 1.23.  In NT we can
                      use the input desktop before anyone logs in.
   0.49       6/1/97  To make it easier to run billgPC where disk images are
                      downloaded from a single source, we now randomize the
                      temporary IP address based on the MAC address when we
                      encounter DHCP.  For the same reason, we'll do this
                      without even asking if both /f and /r are used on the
                      command line.  To make it easier to defeat IP address
                      conflicts that MSTCP reports to the screen but not to
                      WinSock programs, when /f and /r are used and the bootp
                      transaction fails we randomize the IP address and reboot
                      unless the IP address already had the (pseudo-)random
                      value.
   0.50      7/11/97  We now work with the *finished* dynamic VxD.  Fixed an
                      NT-specific bug in using DefineDosDevice().  Found a
                      way around reboot failure in boot-time service-launched
                      billgPC in NT 4.0.  We now can put a message box and
                      the main window on a visible desktop in NT no matter how
                      the program gets launched and when the user logs on.
                      For convenience in debugging we introduce the "/a"
                      command line option to append this run's log to the
                      previous one.
   0.51      11/6/97  To accommodate an NT driver or the 95 VxD with the same
                      app, we needed to move SHARED.H up from the VxD
                      directory and reorganize the IOCTL codes.  Fixed bug
                      in WorkerParams::stackIsMicrosoft() with help from
                      Uwe Proemm.  Fixed C++ archaism in keyNameTail() with
                      help from Andrew Tucker and allowed ignoring of server's
                      hostname with "/xh" option.  Fixed VxD bug as of 1.26
                      level of VxD.
   0.52     12/14/97  We now use the NT driver where possible.  Also we've
                      used Mark Wignall's suggested fix for a bug that
                      afflicted the NetBIOS MAC address detection code.
                      We now can deal with NT 5.  With the NT driver we don't
                      randomize the IP address because we don't need Winsock.
                      We now permit aliasing of RFC 1084 tags so old bootp
                      servers can be tricked into sending new tags.
   0.53      4/28/98  We now change the computer's name using the API call.
                      Fixed a bug causing trouble when billgPC's called from
                      the NT command line.  Before showing the main window
                      we gray the control menu's Close item if we're using
                      the NT driver so we can be sure we've no reads hanging
                      when we depart; we don't do this if the user has
                      chosen to dump the driver or read promiscuously.
                      Made it possible to tell the difference between DNS
                      and NetBIOS name server options in the log.
                      We are now Y2K-correct!  The NT driver has been tested
                      on token ring HW.  When bind() fails with
                      WSAEADDRNOTAVAIL, we fail silently now.
   0.54      10/3/98  Case-insensitive comparison on computer name now, thanks
                      to Mikael Willberg.  We now log our results even when
                      we reboot.  We now retry the NT StartService() for our
                      driver when the service database is locked until the
                      user stops us.  Windows 98 conformity: We no longer
                      assume the ...\UuidTemporaryData key exists; we don't
                      treat the Virtual Private Networking Adapter as "one of
                      our own".  We now handle NetBIOS node type.  Fixed bug
                      in which we failed to use the RFC 1542 (big) size of
                      the vendor options area when needed on RX.  We shun the
                      Toshiba Infrared adapter in Win9X.
*/

#if !defined(_MT) // Symantec seems to need this to believe we're multithreaded
//#define _MT
#endif

#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <process.h>
#include <wincon.h>
#include <ctype.h>
#include <limits.h>
#include <stdio.h>              // for sscanf; sorry!
#include <iostream.h>
#include <strstrea.h>
#include <fstream.h>
#include <iomanip.h>
#include <string.h>
#include <stdlib.h>
#include <winioctl.h>
#include <winsock.h>
#include <nb30.h>
#include "shared.h"
#include "resource.h"

#define VERSION "0.54"
#define TEST_VERSION 0          // if nonzero, annoying messagebox

enum {
  PortBootpS = 67,            // Protocol port for server
  PortBootpC = 68             // Protocol port for client
};

/*
  Once we have an RFC 951-compliant reply, i.e., one that tells us our host IP
  address, the interesting question becomes whether we also have the so-called
  "vendor-specific" info called out in RFC 1084 (which supersedes RFC 1048.)
  Parsing this stuff gets buried in the RFC1084Iterator class, whose purpose,
  once instantiated with an argument pointing to the reply buffer, is to:
  (1) feed the caller a pointer to the current field, if there (still) is one,
      when more() is called;
  (2) find the next field when next() is called.
  
  We also need a data structure for the field itself and an enumerator covering
  the values the fields' identifying tags can take.

  We also keep as a static member a translate table for tags that lets net
  administrators fool their bootp servers into e.g. sending WINS server
  addresses out as cookie server addresses.
*/

struct RFC1084Field {
  UCHAR tag;
  UCHAR length;
  UCHAR data[62]; // much shorter in reality!
  static UCHAR tagXlateTbl[UCHAR_MAX + 1];
};

UCHAR RFC1084Field::tagXlateTbl[256] = { 0 };

enum RFC1084Tag {
  Pad = 0,
  SubnetMask = 1,
  TimeOffset = 2,
  Gateways = 3,
  TimeServers = 4,
  ObsoleteNameServers = 5,
  DomainNameServers = 6,
  LogServers = 7,
  CookieServers = 8,
  LPRServers = 9,
  ImpressServer = 10,
  ResourceLocationServer = 11,
  Hostname = 12,
  BootFileSize = 13,
  MeritDumpFile = 14,
  DomainName = 15,              // actually from RFC 1533
  SwapServer = 16,
  RootPath = 17,
  ExtensionsPath = 18,
  NetBIOSNameServer = 44,
  NetBIOSNodeType = 46,
  MaxOptionSeen = 63,           // This one's from a Novell server
  ReservedMinimum = 128,        // Start of undefined tags
  End = 255
};

struct BigBootpBuf : public BootpBuf { // What RFC 1542 gives us
  BYTE moreVendorArea[1200];
};

enum { RFC1542VendSize = sizeof(BigBootpBuf) - sizeof(BootpBuf) +
       RFC951VendSize
};

class RFC1084Iterator {         // Holds our position in the vendor area
  UCHAR* vend;  // -> abuf.vend
  size_t vendIndex; // points into vend
  void eatPads();
 public:
  RFC1084Iterator(BootpBuf& abuf);
  RFC1084Field* more();
  void next();
};

void RFC1084Iterator::eatPads() // Skip over junk inserted for alignment
{
  while((vend[vendIndex] == UCHAR(Pad)) && (vendIndex < RFC1542VendSize)) {
    vendIndex++;
  }
}

RFC1084Iterator::RFC1084Iterator(BootpBuf& abuf)
: vend(abuf.vend) 
{
  if (ntohl(*((DWORD*)vend)) != RFC1084VendorMagicCookie) {
    vendIndex = RFC1542VendSize; // Make more() return NULL
  }
  else {
    vendIndex = 4;
    eatPads();
  }
}
 
RFC1084Field* RFC1084Iterator::more() // Is there a field left?
{
  if (vendIndex >= RFC1542VendSize) return NULL;
  if (RFC1084Tag(vend[vendIndex]) == End) return NULL;
  if (RFC1084Tag(vend[vendIndex]) >= ReservedMinimum) return NULL;
  return ((RFC1084Field*)(vend + vendIndex));
}
 
void RFC1084Iterator::next()    // On to the next field, if any
{
  if (!more()) return;
  RFC1084Field* f = (RFC1084Field*) (vend + vendIndex);
  vendIndex += (f->length + 2);
  eatPads();
}

/*
  What OS are we running on?  This will influence the type of message box
  we're likely to put up should we need the user's input.
*/

enum OSType { W95, WNT };

OSType os = W95;
DWORD osMajorVersion, osMinorVersion;

UINT mbType = MB_APPLMODAL;

HDESK mainWindowDesktop = NULL; // used in NT desktop switching

/*
  Navigating the registry:
*/

// Table listing info on legal root keys:

struct { 
  char* name;
  HKEY handle;
} rootKeys[] = {
  { "HKEY_CLASSES_ROOT", HKEY_CLASSES_ROOT },
  { "HKEY_CURRENT_USER", HKEY_CURRENT_USER },
  { "HKEY_LOCAL_MACHINE", HKEY_LOCAL_MACHINE },
  { "HKEY_USERS", HKEY_USERS },
  { NULL, NULL }
};

/*
   These are the names of the registry keys we use the most:
*/

char* tcpKeyName = "";          // filled in by configureForOS()
char* netTransKeyName = "HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\"
                        "Services\\Class\\NetTrans"; // Win95

/*
   We'll often need to know if a key is a root key because if so we'd better
   not close its handle.
*/

BOOL isRootKey(HKEY h)
{
  for (int i = 0; rootKeys[i].name != NULL; i++) {
    if (rootKeys[i].handle == h) return TRUE;
  }
  return FALSE;
}

/*
   It's impossible to get a handle for a non-root key; you have to start
   with its root key and work outward.
*/

HKEY getRootKey(const char* apath)
{
  for (int i = 0; rootKeys[i].name != NULL; i++) {
    size_t rkiLen = strlen(rootKeys[i].name);
    if (!strncmp(apath, rootKeys[i].name, rkiLen)) {
      if ((apath[rkiLen] == '\0') || (apath[rkiLen] == '\\')) {
        return rootKeys[i].handle;
      }
    }
  }
  return NULL;
}

ofstream outFile;

enum {
  logLineLen = 500,             // Use this in ostrstreams for listbox string.
  msgLen = 500,                 // Use this for message boxes.
  ObjectNameLen = logLineLen / 2 // Use this for desktop names.
};

/*
  If we have two handles to NT desktops, the objects may be the same even
  if the handles are different.  That's what we test here.
*/

BOOL differentObjects(HANDLE ahandleA, HANDLE ahandleB)
{
  BOOL result = FALSE;
  char objectNameA[ObjectNameLen];
  char objectNameB[ObjectNameLen];
  DWORD ignoreLenNeeded;
  if ((GetUserObjectInformation(ahandleA, UOI_NAME, objectNameA,
                                ObjectNameLen, &ignoreLenNeeded)) &&
      (GetUserObjectInformation(ahandleB, UOI_NAME, objectNameB,
                                ObjectNameLen, &ignoreLenNeeded))) {
    result = strcmp(objectNameA, objectNameB);
  }
  else outFile << "Can't get object name" << GetLastError() << endl;
  return result;
}

/*
   Send the main window's listbox one line of text.  After we send the line
   we select it to make sure it's visible, and then we deselect it so it
   won't call attention to itself.  If we're in a desperate hurry, though -
   dump() is, due to its responsibility in promiscuous mode - we do without
   the visibility manipulation.  By the way, while this function would run
   faster using PostMessage() than with SendMessage(), the data wouldn't get
   through reliably.

   There's a kink here that was introduced when we made it possible to run,
   launched by an NT service at boot time, trying to grab a desktop we wouldn't
   normally have (WinLogon.)  Since we can't execute SetThreadDesktop once
   we already have a window, we need the ability to log messages without a
   main window, i.e. straight to billgpc.log.
*/

void logLine(HANDLE amainWindow, char* aline, BOOL ahurry = FALSE)
{
  if (!amainWindow) {
    outFile << aline << endl;
    return;
  }
  static HWND dlg = INVALID_HANDLE_VALUE;
  if (dlg == INVALID_HANDLE_VALUE) dlg = GetDlgItem(amainWindow, IDC_LOGBOX);
  SendMessage(dlg, LB_ADDSTRING, 0, (LPARAM)aline);
  if (!ahurry) {
    LONG lastLine = SendMessage(dlg, LB_GETCOUNT, 0, 0) - 1;
    SendMessage(dlg, LB_SETCARETINDEX, WPARAM(lastLine), MAKELPARAM(FALSE, 0));
    UpdateWindow(dlg);
  }
}

/*
  This is a convenience function for the cases where we just have a string
  to put out together with Windows's wisdom on what might have happened.
  Just to be tidy, we strip the trailing CRLF in the Windows error message.
*/

void logLastError(HANDLE amainWindow, char* aline)
{
  char* errMsgBuf = "";
  char* msgBuf = new char[logLineLen];
  ostrstream msgOss(msgBuf, logLineLen);
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                NULL, GetLastError(),
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                (LPTSTR) &errMsgBuf, 0, NULL);
  char* crLoc = strchr(errMsgBuf, '\r');
  if (crLoc) *crLoc = '\0';
  msgOss << aline << ":  " << errMsgBuf << ends;
  logLine(amainWindow, msgBuf);
  LocalFree(errMsgBuf);
}

/*
  Running under NT, launched by a service, things get complicated.  Originally
  our aproach involved a window of 0 to go along with our
  MB_APPLMODAL | MB_SERVICE_NOTIFICATION style.  But we like to make the
  whole context of the bootp transaction, i.e. the log in our main window,
  available when things get bad enough for us to put up a message box.  So
  we devised a scheme that allowed us to switch desktops temporarily back
  to the one our main window lives in, and then restore the current input
  desktop once the message box is dismissed.  If this seems bizarre, consider
  that the situation really does arise when the following events
  transpire:
  
  - billgPC is launched by its service before anyone logs on and puts its
  main window where it can be seen, i.e. in the Winlogon desktop;

  - someone logs on before billgPC finishes, so the Default desktop becomes
  current and billgPC's main window becomes invisible;

  - billgPC learns of something that requires the user's attention and wants
  to put up a message box anchored by the main window.

  By the way, there's plenty of time for the second event in this sequence
  to interpose itself between the first and third when the third is a bootp
  failure, i.e. a timeout.
*/

int ourMessageBox(HWND hwnd, char* amsg, UINT atype)
{
#if 0                           // This puts messagebox under window!
  HWND wnd = (os == W95) ? hwnd : 0;
  return MessageBox(wnd, amsg, "billgPC", mbType | atype);
#elif 1
  int result;
  ShowWindow(hwnd, SW_RESTORE);
  if (os == WNT) {
    HDESK newDesktop = OpenInputDesktop(0, FALSE, DESKTOP_SWITCHDESKTOP);
    if (!newDesktop) logLastError(hwnd, "Can't open input desktop");
    if (differentObjects(mainWindowDesktop,
                         GetThreadDesktop(GetCurrentThreadId()))) {
      if (!SetThreadDesktop(mainWindowDesktop)) {
        logLastError(hwnd, "SetThreadDesktop for main window");
      }
    }
    if ((mainWindowDesktop && newDesktop) &&
        differentObjects(mainWindowDesktop, newDesktop)) {
      if (SwitchDesktop(mainWindowDesktop)) {
        result = MessageBox(hwnd, amsg, "billgPC", atype);
        if (!result) logLastError(hwnd, "MessageBox");
        if (!SwitchDesktop(newDesktop)) {
          logLastError(hwnd, "Couldn't switch back to new desktop");
        }
        return result;
      }
      else logLastError(hwnd, "Couldn't switch to old desktop");
    } // else fall through
  } // end of NT-specific logic
  result = MessageBox(hwnd, amsg, "billgPC", atype);
  if (!result) logLastError(hwnd, "MessageBox");
  return result;
#else
  return MessageBox(hwnd, amsg, "billgPC", atype);
#endif
}

/*
   Two functions, alert() and fail(), centralize our error messages
   and ensure that when an error message goes up on the screen the listbox
   is visible so the user has some context for the message.  A third,
   yesOrNo(), does something similar when the user must decide.  These
   functions also make sure messages and responses get logged.
*/

/*
   Here we've reached a point where we can't go on.  This is a convenience
   function removing some Windows GUI clutter.  For those cases in which
   we think it's OK without alarming the user, there's an optional Boolean
   allowing this.
*/

void fail(HWND amainWindow, char* amsg, BOOL asilent = FALSE)
{
  if (!asilent) ourMessageBox(amainWindow, amsg, MB_ICONSTOP);
  logLine(amainWindow, "**Fatal error:");
  logLine(amainWindow, amsg);
  if (amainWindow) SendMessage(amainWindow, WM_CLOSE, 0, 0);
  else exit(-1);
}

/*
   Here we need to alert the user to something that might not be fatal.
*/

void alert(HWND amainWindow, char* amsg)
{
  ourMessageBox(amainWindow, amsg, MB_ICONEXCLAMATION);
  logLine(amainWindow, amsg);
}

/*
   Here we need to ask the user a yes-or-no question.
*/

int yesOrNo(HWND amainWindow, char* amsg)
{
  int result;
  result = ourMessageBox(amainWindow, amsg, MB_YESNO);
  logLine(amainWindow, "**Query:");
  logLine(amainWindow, amsg);
  logLine(amainWindow, ((result == IDYES) ? "[yes]" : "[no]"));
  return result;
}

/*
  Here, since we know we're running on NT, we do what's necessary to make
  ourselves useful should we be running launched by a service before logon.
  (This is after all the best way to run billgPC on NT.)  We see if we're
  already running on the input desktop and, if not, try to seize it.  If
  we do switch desktops we make sure our window won't be minimized, because
  in our experience a minimized window on the NT Winlogon desktop is a dead
  duck.
*/

void setNTDesktop(int& acmdShow, const OSVERSIONINFO& avInfo)
{
  char* msgBuf = new char[logLineLen];
  ostrstream msgOss(msgBuf, logLineLen);
  mbType = (avInfo.dwMajorVersion < 4) ? // Service notification
    MB_TASKMODAL | 0x00040000 : MB_TASKMODAL | 0x00200000;
  enum { ObjectNameLen = logLineLen / 2 };
  char oldDesktopName[ObjectNameLen];
  char newDesktopName[ObjectNameLen];
  DWORD ignoreLenNeeded;
  HWINSTA windowStation = GetProcessWindowStation();
  if (windowStation) {
    char stationName[ObjectNameLen];
    if (GetUserObjectInformation(windowStation, UOI_NAME, stationName,
                                 ObjectNameLen, &ignoreLenNeeded)) {
      msgOss << "Window station name: " << stationName << ends;
      logLine(0, msgBuf);
      msgOss.seekp(0);
    }
    else logLastError(0, "Can't get window station name");
  }
  else logLastError(0, "Can't get window station");
  HDESK oldDesktop = GetThreadDesktop(GetCurrentThreadId());
  if (oldDesktop) {
    if (GetUserObjectInformation(oldDesktop, UOI_NAME, oldDesktopName,
                                 ObjectNameLen, &ignoreLenNeeded)) {
      msgOss << "Old desktop name: " << oldDesktopName << ends;
      logLine(0, msgBuf);
      msgOss.seekp(0);
    }
    else logLastError(0, "Can't get old desktop name");
  }
  else logLastError(0, "Can't get thread desktop");
  HDESK newDesktop = OpenInputDesktop(0, TRUE, DESKTOP_CREATEWINDOW |
                                      DESKTOP_SWITCHDESKTOP);
  if (newDesktop) {
    if (GetUserObjectInformation(newDesktop, UOI_NAME, newDesktopName,
                                 ObjectNameLen, &ignoreLenNeeded)) {
      msgOss << "New desktop name: " << newDesktopName << ends;
      logLine(0, msgBuf);
      msgOss.seekp(0);
    }
    else logLastError(0, "Can't get new desktop name");
    if (strcmp(oldDesktopName, newDesktopName)) {
      if (SetThreadDesktop(newDesktop)){
        acmdShow = SW_SHOWNORMAL;
        mainWindowDesktop = newDesktop;
      }
      else logLastError(0, "SetThreadDesktop");
    }
    else {
      mainWindowDesktop = oldDesktop;
      logLine(0, "Already have the input desktop");
    }
  }
  else logLastError(0, "Can't open input desktop");
}

/*
  The handles for dealing with the NT billgPS driver:
*/

SC_HANDLE scmHandle = NULL;
SC_HANDLE srvHandle = NULL;
HANDLE ntBgpsHandle = INVALID_HANDLE_VALUE;
HANDLE ourInstance = INVALID_HANDLE_VALUE;

/*
  Normally the driver doesn't already exist as a service when we're called,
  but if it already exists that's OK.  We know the current directory has been
  set to the directory billgPC was loaded from.  The call to GetFileAttributes
  is necessary because CreateService will vacuously succeed even if it fails
  to find the driver.
*/

BOOL installNTDriver(SC_HANDLE ascmHandle)
{
  BOOL result = FALSE;
  char driverPath[logLineLen];
  *driverPath = 0;
  GetCurrentDirectory(logLineLen, driverPath);
  wsprintf(driverPath + strlen(driverPath), "\\billgPS.SYS");

  if (GetFileAttributes(driverPath) != 0xffffffff) {
    SC_HANDLE srvHandle =
      CreateService(ascmHandle, "BILLGPS", "BOOTP Support",
                    SERVICE_ALL_ACCESS,
                    SERVICE_KERNEL_DRIVER,
                    SERVICE_DEMAND_START,
                    SERVICE_ERROR_NORMAL,
                    driverPath,
                    NULL, NULL, NULL, NULL, NULL);
    if (srvHandle == NULL) {
      if (GetLastError() == ERROR_SERVICE_EXISTS) {
        logLine(0, "billgPS.SYS already existed");
        result = TRUE;
      }
      else logLastError(0, "Couldn't install billgPS.SYS");
    }
    else {
      logLine(0, "Created service for billgPS.SYS");
      result = TRUE;
    }
  }
  else logLine(0, "Can't find billgPS.SYS");
  if (result == TRUE) CloseServiceHandle(srvHandle);
  return result;
}

/*
  Here we wait for success (in which case we return TRUE) or either of two
  kinds of failure:
  - we get an error from StartService() other than an indication that the
    service database is still locked;
  - the user gives up.
  In either error condition we log an error message and return FALSE.
*/

BOOL APIENTRY startServiceDialogProc(HWND hDlg, UINT message, UINT wParam,
                                     LONG /* lParam */)
{
  switch (message)
    {
    case WM_INITDIALOG:
      {
        SetDlgItemText(hDlg, IDC_EDIT_WAIT_SCM,
                       "Windows NT has reported to billgPC that the "
                       "service database is locked.  This means that "
                       "billgPC's device driver cannot yet be started. "
                       "Normally this is "
                       "a temporary situation, especially when billgPC "
                       "is being used to configure a new workstation.\r\n"
                       "Unless you press the button below, billgPC will "
                       "continue trying to start its driver until it "
                       "succeeds.\r\n"
                       "If you press the button, billgPC will try to reach "
                       "the bootp server without the use of its device "
                       "driver.  You are strongly urged to wait.");
        SetTimer(hDlg, 0, 5000, NULL);
      }
      break;
    case WM_COMMAND:
      if (wParam == IDC_BUTTON_STOP) {
        SetLastError(ERROR_SERVICE_DATABASE_LOCKED);
        logLastError(0, "Couldn't start billgPS service");
        return EndDialog(hDlg, FALSE);
      }
      break;
    case WM_TIMER:
      if (StartService(srvHandle, 0, NULL)) return EndDialog (hDlg, TRUE);
      else {
        DWORD error = GetLastError();
        if (error != ERROR_SERVICE_DATABASE_LOCKED) {
          SetLastError(error);
          logLastError(0, "Couldn't start billgPS service");
          return EndDialog (hDlg, FALSE);
        }
      }
      break;
    }
  return FALSE;
}

/*
  Because our call to StartService() can be frustrated by a transitory lock
  on the service database held by some other process, we keep trying to
  start our driver periodically while displaying a dialog that allows the
  user to give up.
*/

void ourStartService()
{
  BOOL result = FALSE;
  if (StartService(srvHandle, 0, NULL)) result = TRUE;
  else {
    DWORD error = GetLastError();
    if (error == ERROR_SERVICE_DATABASE_LOCKED) {
      result =  DialogBox(ourInstance, MAKEINTRESOURCE(DIALOG_WAIT_SCM), 0,
                          (DLGPROC)startServiceDialogProc);
    }
    else {
      SetLastError(error);
      logLastError(0, "Couldn't start billgPS service");
    }
  }
  if (result) logLine(0, "Started billgPS service");
}

void connectToNTDriver()
{
  scmHandle = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
  if (scmHandle == NULL) fail(0, "OpenSCManager");
  if (installNTDriver(scmHandle)) {
    srvHandle = OpenService(scmHandle, "billgPS", SERVICE_ALL_ACCESS);
    if (srvHandle != NULL) ourStartService();
    else logLastError(0, "Couldn't open billgPS service");
  }
}

void disconnectFromNTDriver()
{
  if (ntBgpsHandle != INVALID_HANDLE_VALUE) {
    if (!CloseHandle(ntBgpsHandle)) {
      logLastError(0, "Can't close billgPS device");
    }
  }
  if (srvHandle != NULL) {
    SERVICE_STATUS  serviceStatus;
    SC_HANDLE stopHandle = OpenService(scmHandle, "billgPS",
                                       SERVICE_ALL_ACCESS);
    if (stopHandle == NULL) logLastError(0, "OpenService stopping billgPS");
    else if (!ControlService(stopHandle, SERVICE_CONTROL_STOP,
                             &serviceStatus)) {
      logLastError(0, "ControlService SERVICE_STOP");
    }
    CloseServiceHandle(srvHandle);
    SC_HANDLE removeHandle = OpenService(scmHandle, "billgPS",
                                         SERVICE_ALL_ACCESS);
    if (removeHandle == NULL) logLastError(0, "OpenService removing billgPS");
    else {
      if (!DeleteService(removeHandle)) {
        logLastError(0, "DeleteService for billgPS");
      }
      CloseServiceHandle(removeHandle);
    }
    CloseServiceHandle(srvHandle);
  }
  if (scmHandle != NULL) CloseServiceHandle(scmHandle);
}

void shutdown()
{
  if (os == WNT) disconnectFromNTDriver();
  char dateStr[80], timeStr[80];
  GetDateFormat(LOCALE_USER_DEFAULT, 0, NULL,
                "yyyy'-'MM'-'dd", dateStr, sizeof(dateStr));
  GetTimeFormat(LOCALE_USER_DEFAULT, TIME_FORCE24HOURFORMAT, NULL,
                "HH':'mm':'ss", timeStr, sizeof(timeStr));
  outFile << "billgPC finished " << dateStr << " " << timeStr << endl;
  outFile.close();
}

BOOL ctrlHandler(DWORD actrlChar)
{
  switch(actrlChar)
    {
    case CTRL_SHUTDOWN_EVENT:
    case CTRL_LOGOFF_EVENT:
      shutdown();
      return FALSE;             // Quit the program.
    default:
      break;
    }
  return TRUE;                  // Don't quit.
}

/*
  Since we run under two different operating systems, there are some things
  we need to set up depending on which one it is.  We keep an enumerator
  for the OSes we tolerate.
*/

DWORD transportValueType = REG_SZ;
char* subnetMaskString = "IPMask";

void configureForOS(int& acmdShow)
{
  char* msgBuf = new char[logLineLen];
  ostrstream msgOss(msgBuf, logLineLen);
  char* osString = "";
  OSVERSIONINFO vinfo;
  vinfo.dwOSVersionInfoSize = sizeof(vinfo);
  GetVersionEx(&vinfo);
  osMajorVersion = vinfo.dwMajorVersion;
  osMinorVersion = vinfo.dwMinorVersion;
  switch(vinfo.dwPlatformId) {
  case VER_PLATFORM_WIN32_WINDOWS:
    os = W95;
    osString = "95";
    tcpKeyName = "HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\"
      "Services\\VxD\\MSTCP";
    transportValueType = REG_SZ;
    subnetMaskString = "IPMask";
    mbType = MB_APPLMODAL;
    break;
  case VER_PLATFORM_WIN32_NT:
    os = WNT;
    osString = "NT";
    tcpKeyName = "HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\"
      "Services\\Tcpip\\Parameters";
    transportValueType = REG_MULTI_SZ;
    subnetMaskString = "SubnetMask";
    setNTDesktop(acmdShow, vinfo);
#if 0 // Control handler doesn't get called at system shutdown: why??
    if (!SetConsoleCtrlHandler((PHANDLER_ROUTINE)ctrlHandler, TRUE)) {
      fail(0, "Can't install shutdown handler");
    }
#endif
    connectToNTDriver();
    break;
  default:
    fail(0, "This program requires Windows 95 or NT");
    break;
  }
  msgOss << "Running under Windows " << osString << " "
         << osMajorVersion << "." << osMinorVersion << "." << ends;
  logLine(0, msgBuf);
#if TEST_VERSION
  MessageBox(0, "This is a test version of billgPC.\n\n"
             "It is designed to be used for a limited time so that\n"
             "you can test it in a network environment that isn't\n"
             "available to the author.\n\n"
             "When testing is complete please email your results to\n"
             "perin@acm.org\n"
             "and you will be given access to a new version that\n"
             "doesn't put up this annoying reminder.",
             "Testing billgPC...",
             MB_TASKMODAL | MB_SETFOREGROUND | MB_OK);
#endif
}

/*
   Here we can't proceed because a registry key we need doesn't exist.
*/

void abortOnBadKey(const char* apath, HWND amainWindow)
{
  char* msgBuf = new char[logLineLen];
  ostrstream msgOss(msgBuf, logLineLen);
  msgOss << "Can't find key" << endl
    << "  " << apath << endl
    << "in registry.\n"
    << "Make sure Microsoft TCP/IP is properly installed"
    << ((os == W95) ? "." : " and that you have Registry write access for IP.")
    << ends;
  fail(amainWindow, msgBuf);
}

/*
   As we navigate outward from the root to the key we seek, we make sure to
   close all intermediate keys.  While navigating we ask only read access
   but at our destination we want to be able to modify things.
*/

HKEY getKey(const char* apath, HWND amainWindow, BOOL asilent = FALSE)
{
  HKEY thisKeyHandle = getRootKey(apath);
  if (thisKeyHandle == NULL) abortOnBadKey(apath, amainWindow);
  char subKeyName[80];
  const char* separatorP = apath + strcspn(apath, "\\");
  while (*separatorP) {
    const char* nextSubKeyP = separatorP + 1;
    size_t subKeyLen = strcspn(nextSubKeyP, "\\");
    memcpy(subKeyName, nextSubKeyP, subKeyLen);
    subKeyName[subKeyLen] = '\0';
    separatorP = nextSubKeyP + subKeyLen;
    HKEY nextKeyHandle;
    if (RegOpenKeyEx(thisKeyHandle, subKeyName, 0,
                     KEY_READ,
                     &nextKeyHandle)!= ERROR_SUCCESS) {
      if (!asilent) {
        char* msgBuf = new char[logLineLen];
        ostrstream msgOss(msgBuf, logLineLen);
        msgOss << "Failure on " << apath << " at subkey: " << subKeyName  <<
          " with read access." << ends;
        logLine(amainWindow, msgBuf);
      }
      RegCloseKey(thisKeyHandle);
      return NULL;
    }
    RegCloseKey(nextKeyHandle);
    if (RegOpenKeyEx(thisKeyHandle, subKeyName, 0,
                     KEY_READ | KEY_SET_VALUE,
                     &nextKeyHandle)!= ERROR_SUCCESS) {
      if (!asilent) {
        char* msgBuf = new char[logLineLen];
        ostrstream msgOss(msgBuf, logLineLen);
        msgOss << "Failure on " << apath << " at subkey: " << subKeyName  <<
          " with full access." << ends;
        logLine(amainWindow, msgBuf);
      }
      RegCloseKey(thisKeyHandle);
      return NULL;
    }
    if (!isRootKey(thisKeyHandle)) RegCloseKey(thisKeyHandle);
    thisKeyHandle = nextKeyHandle;
  }
  return thisKeyHandle;
}

/*
  Sometimes we just need to know if a Registry key exists...
*/

BOOL keyExists(const char* apath, HWND amainWindow)
{
  BOOL result = FALSE;
  HKEY keyHandle = getKey(apath, amainWindow, TRUE);
  if (keyHandle != NULL) {
    RegCloseKey(keyHandle);
    result = TRUE;
  }
  return result;
}

/*
   If we can find a value corresponding to the key name and expected name,
   return TRUE and stow its length.  The argument "alength" is used a la
   Microsoft, i.e. both as input and output.  We don't leave a non-root
   key handle open.
*/

BOOL getValue(const char* akeyName, char* anexpectedName, BYTE* abuf,
              DWORD& alength, DWORD atype, HWND amainWindow)
{
  HKEY keyHandle = getKey(akeyName, amainWindow);
  BOOL result = FALSE;
  DWORD regType;                // value's type
  if (keyHandle == NULL) abortOnBadKey(akeyName, amainWindow);
  if (RegQueryValueEx(keyHandle, anexpectedName, NULL,
                      &regType, abuf, &alength) == ERROR_SUCCESS) {
    if (regType == atype) result = TRUE;
    else {
      size_t bufLen = strlen(anexpectedName) + strlen(akeyName) + 100;
      char* msgBuf = new char[bufLen];
      ostrstream msgOss(msgBuf, bufLen);
      msgOss << "Unexpected value type in registry for" << endl
             << "  key: " << akeyName << endl
             << "  value name: " << anexpectedName << ends;
      alert(amainWindow, msgBuf);
    }
  }
  if (!isRootKey(keyHandle)) RegCloseKey(keyHandle);
  return result;
}

/* If we're reduced to using NetBIOS to get the MAC address, we try as
  many LANAs (to use NetBIOS slang) as we need to find a netcard we
  can deal with.  For each LANA, before we issue the NCBASTAT that may
  give us the info we need, we must issue an NCBRESET.  We also set
  the bootp hardware type (not to be confused with the NetBIOS adapter
  type) while we're at it.
*/

BOOL getNetBIOSHardwareAddress(BYTE* abuf, DWORD& alength, BYTE& ahwType)
{
  enum { NBEther = 0xfe, NBToken = 0xff }; // adapter types we love
  enum { NNames = 10 };
  struct {
    ADAPTER_STATUS adapt;
    NAME_BUFFER nameBuff[NNames];
  } asBuf;
  NCB ncb;
  LANA_ENUM lanaEnum;
  BOOL result = FALSE;
  
  memset(&ncb, '\0', sizeof(ncb));
  ncb.ncb_command = NCBENUM;
  ncb.ncb_buffer = (UCHAR*)&lanaEnum;
  ncb.ncb_length = sizeof(lanaEnum);
  Netbios(&ncb);
  if (ncb.ncb_retcode == NRC_GOODRET) {
    for (UCHAR lanaNo = 0; !result && (lanaNo < lanaEnum.length); lanaNo++) {
      memset(&ncb, '\0', sizeof(ncb));
      ncb.ncb_buffer = (UCHAR*)&asBuf;
#if 0
      ncb.ncb_lana_num = lanaNo;
#else  // Thanks, Mark Wignall!
      ncb.ncb_lana_num = lanaEnum.lana[lanaNo];
#endif
      memset(ncb.ncb_callname, 0, NCBNAMSZ);
      ncb.ncb_length = sizeof(asBuf);
      ncb.ncb_callname[0] = 1;
      ncb.ncb_callname[2] = NNames;
      ncb.ncb_command = NCBRESET;
      Netbios(&ncb);
      memset(ncb.ncb_callname, ' ', NCBNAMSZ);
      *ncb.ncb_callname = '*';
      ncb.ncb_length = sizeof(asBuf);
      ncb.ncb_command = NCBASTAT;
      Netbios(&ncb);
      if (((ncb.ncb_retcode == NRC_GOODRET) ||
           (ncb.ncb_retcode == NRC_INCOMP)) &&
          ((asBuf.adapt.adapter_type == NBEther) ||
           (asBuf.adapt.adapter_type == NBToken))) {
        alength = sizeof(asBuf.adapt.adapter_address);
        memcpy(abuf, asBuf.adapt.adapter_address, alength);
        if (asBuf.adapt.adapter_type == UCHAR(NBEther)) {
          ahwType = BootpHWTypeEthernet;
        }
        else ahwType = BootpHWTypeIEEE;
        result = TRUE;
      }
    }
  }
  return result;
}

/*
  Shut down a socket and, if asked, dissolve our relationship to
  WinSock.  Note well that this alerts any thread using this socket
  to the effect that it isn't valid any more.

  NT's incontinence about sockets not closed by a dying app makes it vital
  that this function be executed by the WorkerParams destructor.
*/

void quiesceSocket(SOCKET& asock, BOOL acleanup)
{
  if (asock != INVALID_SOCKET) {
    closesocket(asock);
    asock = INVALID_SOCKET;
    if (acleanup) WSACleanup();
  }
}

/*
  Because it makes the program easier to understand, we've factored out the
  sender of bootp requests and the receiver of bootp replies as separate
  threads from the main thread.  This creates a need for data structures
  giving the threads access to the information they need.  We've also
  factored out the relatively long-winded post-receiver stuff into a "worker"
  thread that stays out of the way of the GUI.
*/

struct SenderParams {
  SOCKET& sock;
  BootpBuf* buf;
  HWND mainWindow;
  SenderParams(SOCKET& asock, BootpBuf* abuf, HWND amainWindow) :
    sock(asock), buf(abuf), mainWindow(amainWindow) { }
};

/*
  The timeout that the receiver thread itself uses to quit on a read from
  billgPS is made shorter than the one the worker thread uses to kill the
  receiver thread by the initialization of ReceiverParams.
*/

struct ReceiverParams {
  SOCKET& sock;
  BigBootpBuf* buf;
  DWORD xid;
  HWND mainWindow;
  HANDLE vxdHandle;
  BOOL promiscuous;             // accept first packet, no matter what?
  DWORD timeout;
  BOOL success;
  ReceiverParams(SOCKET& asock, BigBootpBuf* abuf, DWORD axid,
                 HWND amainWindow, HANDLE avxdHandle,
                 BOOL apromiscuous, DWORD atimeout) :
    sock(asock), buf(abuf), xid(axid), mainWindow(amainWindow),
    vxdHandle(avxdHandle), promiscuous(apromiscuous),
    timeout((atimeout > 2000) ? (atimeout - 2000) : 1000 ), success(FALSE) { }
};

struct WorkerParams {
  HWND mainWindow;
  BOOL forceFix;                // force fixing Registry discrepancies?
  BOOL promiscuous;             // accept first packet, no matter what?
  BOOL dumpDriver;              // dump the driver?
  BOOL forceIEEE;               // force IEEE htype in bootp request?
  BOOL forceEthernet;           // force Ethernet htype in bootp request?
  BOOL forceReboot;             // force reboot on Registry change?
  BOOL askBroadcast;            // ask for broadcast reply first time?
  BOOL updateComputerName;      // Hostname coerces ComputerName
  BOOL ignoreHostName;          // Hostname gets ignored
  DWORD rxTimeout;              // how long we wait for reply (msec.)
  DWORD hops;                   // initial hop count
  char* transportKeyName;       // card-specific TCP transport key name
  SOCKET sock;
  char* adapterName;            // e.g. "Elnk31"
  WorkerParams(HWND amainWindow, BOOL aforceFix, BOOL apromiscuous,
               BOOL adumpDriver, BOOL aforceIEEE, BOOL aforceReboot,
               BOOL anaskBroadcast, DWORD arxTimeout, BOOL aforceEthernet,
               BOOL anupdateComputerName, DWORD ahops, BOOL anignoreHostName) :
    mainWindow(amainWindow), forceFix(aforceFix),
    promiscuous(apromiscuous), dumpDriver(adumpDriver),
    forceIEEE(aforceIEEE),forceEthernet(aforceEthernet),
    forceReboot(aforceReboot), updateComputerName(anupdateComputerName),
    askBroadcast(anaskBroadcast), rxTimeout(arxTimeout), hops(ahops),
    ignoreHostName(anignoreHostName),
    transportKeyName(NULL), sock(INVALID_SOCKET), adapterName(NULL) { };
  ~WorkerParams() {
    if (transportKeyName) delete transportKeyName;
    if (adapterName) delete adapterName;
    quiesceSocket(sock, TRUE);
  };
  UINT changeComputerName(char* anewName);
  BOOL doReboot();
  void dumpDriverMemory(HANDLE avxdHandle);
  char* findNT4NetcardTCPParams();
  char* findNT5NetcardTCPParams();
  char* findTCPNetTransKeyName();
  BOOL stackIsMicrosoft();
  BOOL isDHCP();
  DWORD fillSendBuf(BootpBuf& asendBuf, HANDLE avxdHandle);
  BOOL grabRebootPrivilege();
  void logServerAddress(UCHAR* anaddr);
  void openNTBgps();
  UINT registerResponse(BootpBuf& buf);
  UINT regify(const char* akeyName, DWORD atype, char* atitle,
              const void* abootpValue, BOOL aforceUpdate,
              const char* alogPrefix = NULL);
  UINT regIPAddrs(const char* akeyName, DWORD atype, char* atitle,
                  const UCHAR* anaddr, int alen,
                  int alimit = sizeof(BootpBuf),
                  const char* alogPrefix = NULL);
  UINT regHostname(RFC1084Field* f);
  UINT regDomainName(RFC1084Field* f);
  UINT regNetBIOSNameServer(RFC1084Field* f);
  UINT regNetBIOSNodeType(RFC1084Field* f);
  SOCKET getBootpSocket();
  BOOL getNTHardwareAddress(BYTE* abuf, DWORD& alength, BYTE& ahwType);
  BOOL getNtBgpsHardwareAddress(BYTE* abuf, DWORD& alength, BYTE& ahwType);
  void getHardwareAddress(BYTE* abuf, DWORD& alength, BYTE& ahwType,
                          HANDLE avxdHandle, BOOL alog);
  void setupBootp(HANDLE avxdHandle);
  void checkStack(HANDLE avxdHandle);
  void allowReboot();
  void tryReboot();
  BOOL disconnected();
  DWORD valueLen(const void* avalue, DWORD atype);
  void handleBootpFailure(HANDLE avxdHandle);
  UINT forcePseudoRandomIPAddress(HANDLE avxdHandle);
};

/*
  Here we use our driver's facilities for querying the hardware.
*/

BOOL WorkerParams::getNtBgpsHardwareAddress(BYTE* abuf, DWORD& alength,
                                            BYTE& ahwType)
{
  char allZeroes[6] = { 0 };
  DWORD length;
  BYTE hwType;
  BOOL result = FALSE;
  if ((ntBgpsHandle != INVALID_HANDLE_VALUE) && // only in Win95
      (DeviceIoControl(ntBgpsHandle, DIOC_BillgpGetMacAddress,
                       NULL, 0, abuf, MacAddressSize, &length, NULL)) &&
      memcmp(abuf, allZeroes, sizeof(allZeroes))) {
    DWORD hwLength;
    if (DeviceIoControl(ntBgpsHandle, DIOC_BillgpGetHWType,
                        NULL, 0, &hwType, 1, &hwLength, NULL)) {
      result = TRUE;
      alength = length;
      ahwType = hwType;
    }
  }
  return result;
}

/*
  Here we use a MAC address sniffing technique that works only for NT.
*/

BOOL WorkerParams::getNTHardwareAddress(BYTE* abuf, DWORD& alength,
                                        BYTE& ahwType)
{
  if (os != WNT) return FALSE;
  char queryResult[512];
  char deviceName[80];
  BOOL result = FALSE;
  BOOL createdDevice = FALSE;   // resorted to DefineDosDevice?
  BOOL foundDevice = BOOL(QueryDosDevice(adapterName, queryResult,
                                         sizeof(queryResult)));
  if ((!foundDevice) && (GetLastError() == ERROR_FILE_NOT_FOUND)) {
    strcpy(deviceName, "\\Device\\");
    strcat(deviceName, adapterName);
    createdDevice = DefineDosDevice(DDD_RAW_TARGET_PATH, adapterName,
                                    deviceName);
  }

  if (foundDevice || createdDevice) {
    char macFileName[80];
    strcpy(macFileName, "\\\\.\\");
    strcat(macFileName, adapterName);
    HANDLE hMAC = CreateFile(macFileName, GENERIC_READ,
                             FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                             OPEN_EXISTING, 0, INVALID_HANDLE_VALUE);
    if (hMAC != INVALID_HANDLE_VALUE) {
      UCHAR addrData[80];
      ULONG ethAddrCode = 0x01010102;
      ULONG trAddrCode = 0x02010102;
      DWORD returnedCount = DWORD(-1); // impossible value
      BOOL ioctlOK = FALSE;       // assume failure
      DWORD queryIoctl = CTL_CODE(FILE_DEVICE_PHYSICAL_NETCARD, 0,
                                  METHOD_OUT_DIRECT, FILE_ANY_ACCESS);
      if (DeviceIoControl(hMAC, queryIoctl /* IOCTL_NDIS_QUERY_GLOBAL_STATS */,
                          &ethAddrCode, sizeof(ethAddrCode), addrData,
                          sizeof(addrData), &returnedCount, NULL)) {
        ahwType = BootpHWTypeEthernet;
        ioctlOK = TRUE;
      }
      else if (DeviceIoControl(hMAC, queryIoctl,
                               &trAddrCode, sizeof(trAddrCode), addrData,
                               sizeof(addrData), &returnedCount, NULL)) {
        ahwType = BootpHWTypeIEEE;
        ioctlOK = TRUE;
      }
      if (ioctlOK && (returnedCount == 6)) {
        alength = returnedCount;
        memcpy(abuf, addrData, 6);
        result = TRUE;
      }
      if (createdDevice) {        // Get rid of it!
        DefineDosDevice(DDD_RAW_TARGET_PATH | DDD_REMOVE_DEFINITION |
                        DDD_EXACT_MATCH_ON_REMOVE, adapterName, deviceName);
      }
      CloseHandle(hMAC);
    }
  }
  return result;
}

/*
  If billgPS knows our LAN card's MAC address, we trust it.  Failing that,
  we prefer to use NetBIOS to get the MAC address, because our Registry
  value approach won't work if the netcard's ever been replaced, but we
  keep the Registry approach as a fallback position.  We also grab the
  hardware type, which may be Ethernet or token ring.

  This is where the option to force IEEE as the hardware type takes effect.
  This option's importance is for two reasons: the VxD can't tell what form
  of addressing's being used, and where NetBIOS is used the results are
  unreliable: in one documented case a token ring card on an IEEE-encapsulated
  net was called Ethernet by NetBIOS!
*/

void WorkerParams::getHardwareAddress(BYTE* abuf, DWORD& alength,
                                      BYTE& ahwType, HANDLE avxdHandle,
                                      BOOL alog)
{
  char* registryHardwareKeyName = "HKEY_LOCAL_MACHINE\\SOFTWARE\\Description"
    "\\Microsoft\\Rpc\\UuidTemporaryData"; // invalid in Win98
  char allZeroes[6] = { 0 };
  ahwType = BootpHWTypeEthernet; // We'll default to Ethernet.
  if ((avxdHandle != INVALID_HANDLE_VALUE) && // only in Win95
      (DeviceIoControl(avxdHandle, DIOC_BillgpGetMacAddress,
                       NULL, 0, abuf, MacAddressSize, &alength, NULL)) &&
      memcmp(abuf, allZeroes, sizeof(allZeroes))) {
    DeviceIoControl(avxdHandle, DIOC_BillgpGetHWType,
                    NULL, 0, &ahwType, 0, NULL, NULL);
    if (alog) logLine(mainWindow, "MAC address taken from billgPS.VxD:");
  }
  else if (getNtBgpsHardwareAddress(abuf, alength, ahwType)) {
    if (alog) logLine(mainWindow, "MAC address taken from billgPS.SYS:");
  }
  else if (getNTHardwareAddress(abuf, alength, ahwType)) {
    if (alog) logLine(mainWindow, "MAC address taken from netcard query:");
  }
  else if (getNetBIOSHardwareAddress(abuf, alength, ahwType)) {
    if (alog) logLine(mainWindow, "MAC address taken from NetBIOS:");
  }
  else if (keyExists(registryHardwareKeyName, mainWindow) &&
           getValue(registryHardwareKeyName,
                    "NetworkAddress", abuf, alength, REG_BINARY, mainWindow)) {
    if (alog) logLine(mainWindow, "MAC address taken from Registry:");
  }
  else fail(mainWindow, "Can't find hardware address");
  if (forceIEEE) ahwType = BootpHWTypeIEEE; // Override detected type!
  if (forceEthernet) ahwType = BootpHWTypeEthernet; // Override detected type!
  if (alog) {
    char* msgBuf = new char[logLineLen];
    ostrstream msgOss(msgBuf, logLineLen);
    msgOss << "  ";
    for (UINT i = 0; i < alength; i++) {
      msgOss << hex << setw(2) << setfill('0') << WORD(abuf[i]) << " ";
    }
    msgOss << ((ahwType == BootpHWTypeIEEE) ? "(IEEE)" : "(Ethernet)");
    msgOss << ends;
    logLine(mainWindow, msgBuf);
  }
}

/*
  Initialize the bootp send buffer:  We extract the hardware address from
  the registry and fill its statutory representation in the buffer, reusing
  its low-order digits as a reasonably random number to serve as the
  transaction ID.  If our command-line options say so we set the broadcast
  bit so the first request will carry it.
*/

DWORD WorkerParams::fillSendBuf(BootpBuf& asendBuf, HANDLE avxdHandle)
{
  DWORD hwAddrLength = RFC951HaddrLen;
  memset(&asendBuf, 0, sizeof(asendBuf));
  getHardwareAddress(asendBuf.chaddr, hwAddrLength, asendBuf.htype,
                     avxdHandle, TRUE);
  asendBuf.op = RFC951BOOTREQUEST;
  if (hwAddrLength != 6) {      // Not Ethernet, sorry!
    fail(mainWindow, "Can't deal with this type of adapter");
  }
  asendBuf.hlen = BYTE(hwAddrLength);
  // Here's how we get a pseudo-random transaction ID:
  // the least significant digits of the hardware address.
  asendBuf.xid = *((DWORD*)(asendBuf.chaddr +
                           hwAddrLength - sizeof(asendBuf.xid)));
  asendBuf.secs = htons(1);     // Is this superstition?
  asendBuf.hops = BYTE(hops);   // Will this fool Novell's IP stack?
  *((DWORD*)asendBuf.vend) = htonl(RFC1084VendorMagicCookie);
  if (askBroadcast) asendBuf.flags = htons(BootpFlagBroadcast);
  return asendBuf.xid;
}

/*
  The bootp broadcaster's job is to keep sending broadcasts at a reasonable
  interval until its thread is destroyed or its socket goes away.
  Once the first broadcast, which asks for a directed datagram reply, has
  gone unanswered, we request the server broadcast its reply.  This is vital
  if we're receiving via Winsock and don't know our IP address.
*/

void bootpSender(void* aparams)
{
  SenderParams& params = *((SenderParams*)aparams);
  sockaddr_in serverAddr;
  memset (&serverAddr, 0, sizeof(serverAddr));
  serverAddr.sin_family = AF_INET;
  //#define JACK 1
#if defined(JACK)
  serverAddr.sin_addr.s_addr = inet_addr("192.100.100.41");
#else
  serverAddr.sin_addr.s_addr = INADDR_BROADCAST;
#endif
  serverAddr.sin_port = htons(PortBootpS);
  WORD timeoutSeconds = WORD((params.buf->xid & 7) + 7); // >=7, < 15
  while(params.sock != INVALID_SOCKET) {
    if (ntBgpsHandle != INVALID_HANDLE_VALUE) {
      OVERLAPPED ovrlp = { 0 };
      DWORD bytesSent;

      if ((ovrlp.hEvent = CreateEvent(0, FALSE, 0, NULL)) == 0) {
        fail(params.mainWindow, "Can't create event for WriteFile");
      }
      else if (!WriteFile(ntBgpsHandle, params.buf, sizeof(BootpBuf),
                          &bytesSent, &ovrlp)) {
        if (GetLastError() == ERROR_IO_PENDING) {
          GetOverlappedResult(ntBgpsHandle, &ovrlp, &bytesSent, TRUE);
        }
        else logLastError(params.mainWindow, "WriteFile");
      }
    }
    else {
      int rc = sendto(params.sock, (char*)(params.buf), sizeof(BootpBuf), 0,
                      (sockaddr*)&serverAddr, sizeof(serverAddr));
      if (rc < sizeof(BootpBuf)) {
        char* msgBuf = new char[logLineLen];
        ostrstream msgOss(msgBuf, logLineLen);
        msgOss << "sendto failed: " << WSAGetLastError() << ends;
        fail(params.mainWindow, msgBuf);
      }
    }
    params.buf->secs = htons(WORD(ntohs(params.buf->secs) + timeoutSeconds));
    Sleep(timeoutSeconds * 1000);
    params.buf->flags = htons(BootpFlagBroadcast);
  }
  _endthread();
}

/*
   If we get something from the netcard that looks fishy, we want to dump it.
*/

void dump(HANDLE amainWindow, UINT asize, BYTE* abufP, char* alegend)
{
  logLine(amainWindow, alegend);
  for (UINT offset = 0; offset < asize; offset += 16) {
    char* msgBuf = new char[logLineLen];
    ostrstream msgOss(msgBuf, logLineLen);
    msgOss << hex << setw(3) << setfill('0') << offset << ":";
    for (UINT incr = 0; incr < 16; incr++) {
      if (offset + incr >= asize) break;
      if ((incr % 4) == 0) msgOss << " ";
      msgOss << hex << setw(2) << setfill('0') << WORD(abufP[offset + incr]);
    }
    msgOss << ends;
    logLine(amainWindow, msgBuf, TRUE);
  }
}

/*
  Here we find the non-dialup adapter and bind our VxD to it using the
  technique Chris Chlap developed to make the Microsoft packet driver
  dynamic.  If we fail, we close the handle and mark it invalid.
*/

void bindToMAC(HANDLE& avxdHandle, HWND amainWindow)
{
  BOOL boundIt = FALSE;         // bound to "real" MAC yet?
  char* netParentKeyName = "HKEY_LOCAL_MACHINE\\System\\"
    "CurrentControlSet\\Services\\Class\\Net";
  HKEY netParentHandle = getKey(netParentKeyName, amainWindow);
  if (netParentHandle == NULL) abortOnBadKey(netParentKeyName, amainWindow);

  for (DWORD n = 0; !boundIt ; n++) {
    char netSubKeyName[80];   // e.g. "0001"
    DWORD netSubKeyNameSize = sizeof(netSubKeyName);
    FILETIME lastUpdate;

    LONG rc = RegEnumKeyEx(netParentHandle, n, netSubKeyName,
                           &netSubKeyNameSize, NULL, NULL, NULL, &lastUpdate);
    if (rc != ERROR_SUCCESS) break; // No more subkeys.

    char netFullKeyName[300];
    strcpy(netFullKeyName, netParentKeyName);
    strcat(netFullKeyName, "\\");
    strcat(netFullKeyName, netSubKeyName);

    BYTE valueBuf[200];
    DWORD valueLen = sizeof(valueBuf);
    if (getValue(netFullKeyName, "DriverDesc", valueBuf, valueLen, REG_SZ,
                 amainWindow)) {
      // This kludge for avoiding the wrong netcard is due to the fact that
      // there's no general way to find which protocol's bound to which
      // adapter in Win9X.
      char* dialupSignature = "Dial";
      char* vpnaSignature = "Microsoft Virtual Private"; // uses NDISWAN
      char* tFIRSignature = "Toshiba FIR"; // Why doesn't this work?
      if (strnicmp((char*)valueBuf, dialupSignature,
                   strlen(dialupSignature)) &&
          strnicmp((char*)valueBuf, vpnaSignature, strlen(vpnaSignature)) &&
          strnicmp((char*)valueBuf, tFIRSignature, strlen(tFIRSignature))) {
        DWORD bytesReturned   = 0;
        OVERLAPPED ovrlp = {0,0,0,0,0};
    
        if ((ovrlp.hEvent = CreateEvent(0, FALSE, 0, NULL)) == 0) {
          fail(amainWindow, "Windows is out of resources.");
        }
        if (DeviceIoControl(avxdHandle, DIOC_BillgpBind,
                            netSubKeyName, strlen(netSubKeyName) + 1,
                            NULL, 0,
                            &bytesReturned, &ovrlp)) {
          GetOverlappedResult(avxdHandle, &ovrlp, &bytesReturned, TRUE);
          boundIt = TRUE;
        }
      }
    }
  }
  RegCloseKey(netParentHandle);
  if (!boundIt) {
    logLine(amainWindow, "Couldn't bind to netcard.");
    CloseHandle(avxdHandle);
    avxdHandle = INVALID_HANDLE_VALUE;
  }
}

/*
  If we're running under NT there's no driver; otherwise we try to connect
  to our VxD and, if we do, make sure the version is what we need.
*/

HANDLE getBillgpsHandle(HWND amainWindow)
{
  if (os == WNT) return INVALID_HANDLE_VALUE;
  char* vxdName = "\\\\.\\BILLGPS.VXD";
  UINT vxdVersion;
  DWORD bytesReturned;
  enum { VXDVersionMin = 127, VXDVersionMax = 127 };
  HANDLE vxdHandle = CreateFile(vxdName, 0, 0, NULL, 0,
                                FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED |
                                FILE_FLAG_DELETE_ON_CLOSE, NULL);
  if (vxdHandle == INVALID_HANDLE_VALUE)  {
    logLine(amainWindow, "No VxD loaded; using WinSock to receive reply");
  }
  else if (!DeviceIoControl(vxdHandle, DIOC_BillgpGetVersion,
                            NULL, 0, &vxdVersion, sizeof(vxdVersion), 
                            &bytesReturned, NULL)) {
    fail(amainWindow, "Can't do DIOC_BillgpGetVersion IOCTL");
  }
  else if ((vxdVersion < VXDVersionMin) ||(vxdVersion > VXDVersionMax)) {
    fail(amainWindow, "Wrong version of VXD");
  }
  else {
    char* msgBuf = new char[logLineLen];
    ostrstream msgOss(msgBuf, logLineLen);
    msgOss << "Using billgPS.VxD version " <<
      (vxdVersion / 100) << "." << setw(2) << setfill('0') <<
      (vxdVersion % 100) << ends;
    logLine(amainWindow, msgBuf);
    if (vxdVersion > 122) bindToMAC(vxdHandle, amainWindow);
  }

  return vxdHandle;
}

/*
  The receiver thread exits when it detects an RFC 951-compliant reply
  (params->success TRUE) or when the socket fails (FALSE.)

  We need to loop because it's possible we'll need to examine more than one
  incoming buffer.  This is because the existence of sloppy bootp daemons
  that broadcast responses forces our VxD to look at responses that are
  really for other hosts.  Couple that with an underachieving netcard driver
  that won't tell our VxD the card's MAC address and, well, you understand
  our suspicion.  We exit the loop, assuming our thread doesn't get killed
  and the OS doesn't reject any of our requests, on any of these conditions:

  - a good packet arrives;

  - a bad packet arrives (we dump it);

  - after awakening from waiting for a VxD event under Win95 we find that
  the socket's no longer valid, which is our way of detecting that we've
  been the target of a CloseHandle(), which under Win95 seems not to kill
  the thread under these circumstances.
*/

void bootpReceiver(void* aparams)
{
  ReceiverParams& params = *((ReceiverParams*)aparams);

  if ((params.vxdHandle == INVALID_HANDLE_VALUE) &&
      (ntBgpsHandle == INVALID_HANDLE_VALUE)) {
    while (1) {
      int rc = recvfrom(params.sock, (char*)(params.buf), sizeof(BigBootpBuf),
                        0, NULL, NULL);
      if (rc == SOCKET_ERROR) {
        DWORD errorCode = WSAGetLastError();
        if (errorCode != WSAEINTR) { // Not a timeout...
          char* msgBuf = new char[logLineLen];
          ostrstream msgOss(msgBuf, logLineLen);
          msgOss << "WinSock error on receive:" << errorCode << ends;
          fail(params.mainWindow, msgBuf);
        }
        break;
      }
      else if ((rc >= sizeof(BootpBuf)) &&
               (rc <= sizeof(BigBootpBuf)) &&
               (params.buf->op == RFC951BOOTREPLY) &&
               (params.buf->xid == params.xid)) {
        params.success = TRUE;
        break;
      }
    }
  }

  else if (ntBgpsHandle != INVALID_HANDLE_VALUE) { // NT billgPS
    OVERLAPPED ovrlp = {0,0,0,0,0};
    DWORD bytesReturned;
    BOOL gotIt = FALSE;
    if ((ovrlp.hEvent = CreateEvent(0, FALSE, 0, NULL)) == 0) {
      fail(params.mainWindow, "Windows is out of resources.");
    }
    else if (ReadFile(ntBgpsHandle, params.buf, sizeof(BigBootpBuf),
                      &bytesReturned, &ovrlp)) gotIt = TRUE;
    else if (GetLastError() == ERROR_IO_PENDING) {
      if (WaitForSingleObject(ovrlp.hEvent, params.timeout) == WAIT_OBJECT_0) {
        GetOverlappedResult(ntBgpsHandle, &ovrlp, &bytesReturned, FALSE);
        gotIt = TRUE;
      }
    }
    else logLastError(params.mainWindow, "Can't read from billgPS");
    if (gotIt &&
        (params.buf->op == RFC951BOOTREPLY) &&
        (params.buf->xid == params.xid)) { // It's for us.
      params.success = TRUE;
    }
  }

  else {                        // We do have a billgPS VxD.
    OVERLAPPED ovrlp = {0,0,0,0,0};
    int command = params.promiscuous ? DIOC_BillgpAwaitAny :
      DIOC_BillgpAwaitBootp;

    // Create event that'll be used by VxD to indicate operation is complete
    if ((ovrlp.hEvent = CreateEvent(0, FALSE, 0, NULL)) == 0) {
      fail(params.mainWindow, "Windows is out of resources.");
    }
    else while (1) {
      // get the address of the buffer structure
      DWORD bytesReturned;
      if (!DeviceIoControl(params.vxdHandle, command,
                           NULL, 0, params.buf, sizeof(*params.buf), 
                           &bytesReturned, &ovrlp)) {
        fail(params.mainWindow, "Can't do DIOC_BillgpAwaitBuf IOCTL");
      }
      else {
        GetOverlappedResult(params.vxdHandle, &ovrlp, &bytesReturned, TRUE);
        if (params.sock == INVALID_SOCKET) _endthread(); // Thanks, Win95!
        if ((bytesReturned >= sizeof(BootpBuf)) &&
            (bytesReturned <= sizeof(BigBootpBuf)) &&
            (params.buf->op == RFC951BOOTREPLY) &&
            (params.promiscuous == FALSE)) {
          if (params.buf->xid == params.xid) { // It's for us.
            params.success = TRUE;
            break;
          } // Else continue looping
        }
        else {                  // Bad response or promiscuous
          dump(params.mainWindow, bytesReturned, (BYTE*)params.buf,
               "Packet returned:");
          if (!params.promiscuous) break;
        }
      }
    }
  }
  _endthread();
}

/*
  Here we place a human-readable version of a Registry value into a buffer.
  Note that the argument for the size of the value is ignored except for
  binary strings.
*/

void dumpRegValue(ostrstream& anoss, DWORD avalueType, const void* avalue,
                  DWORD avalueSize)
{
  const UCHAR* value = (const UCHAR*)avalue;
  switch(avalueType)
    {
    case REG_BINARY:
      {
        for (DWORD i = 0; i < avalueSize - 1; i++) {
          anoss << hex << setw(2) << setfill('0') << WORD(value[i]) << ".";
        }
        anoss << hex << setw(2) << setfill('0') << WORD(value[avalueSize - 1]);
      }
      break;
    case REG_DWORD_BIG_ENDIAN:
      {
        for (int i = 0; i < sizeof(DWORD) - 1; i++) {
          anoss << hex << setw(2) << setfill('0') << WORD(value[i]) << ".";
        }
        anoss << hex << setw(2) << setfill('0') <<
          WORD(value[sizeof(DWORD) - 1]);
      }
      break;
    case REG_DWORD_LITTLE_ENDIAN:
      anoss << hex << *((DWORD*)value);
      break;
    case REG_EXPAND_SZ:
      anoss << (CHAR*)value;
      break;
    case REG_SZ:
      anoss << (CHAR*)value;
      break;
    case REG_LINK:
      anoss << "(REG_LINK)";
      break;
    case REG_MULTI_SZ:
      {
        anoss << (CHAR*)value;
        for (CHAR* thisValue = (CHAR*)value + strlen((CHAR*)value) + 1;
             *thisValue;
             thisValue += (strlen(thisValue) + 1)) {
          anoss << "~" << thisValue; // Use tilde for separator.
        }
      }
      break;
    case REG_NONE:
      anoss << "(REG_NONE)";
      break;
    case REG_RESOURCE_LIST:
      anoss << "(REG_RESOURCE_LIST)";
      break;
    default:
      anoss << "UNKNOWN DATA TYPE (" << avalueType << ")";
      break;
    }
  anoss << dec;
}

/*
  Here we calculate the length of a registry value of a given type.
*/

DWORD WorkerParams::valueLen(const void* avalue, DWORD atype)
{
  switch(atype) {
  case REG_DWORD_BIG_ENDIAN:
  case REG_DWORD_LITTLE_ENDIAN:
    return sizeof(DWORD);       // used to fall through 'til 0.53.6
  case REG_EXPAND_SZ:
  case REG_SZ:
    return (strlen((const CHAR*)avalue) + 1);
  case REG_MULTI_SZ:
    {
      const CHAR* value = (const CHAR*)avalue;
      DWORD lenSoFar = strlen(value) + 1;
      for (; value[lenSoFar]; lenSoFar += (strlen(value + lenSoFar) + 1)) ;
      return (lenSoFar + 1);
    }
  default:
    fail(mainWindow, "Can't deal with registry data type");
    return 0;
  }
}

/*
   Here's where we handle a bootp vendor area field whose corresponding
   registry value is of type REG_SZ (i.e. null-terminated string.),
   REG_MULTI_SZ (more than 1 string) or REG_DWORD_LITTLE_ENDIAN.

   In this and all functions that can write to the registry, we return a
   positive number if we did so, zero otherwise.
*/

enum { RegValueBufLen = 250 };        // room for 15 IP address expansions!

UINT WorkerParams::regify(const char* akeyName, DWORD atype, char* atitle,
                          const void* abootpValue, BOOL aforceUpdate,
                          const char* alogPrefix)
{
  UINT result = 0;
  char registryBuf[RegValueBufLen];
  DWORD registryBufLen = RegValueBufLen;
  *registryBuf = '\0';      // Just in case...
  BOOL discrepancy = FALSE;
  char* logBuf = new char[logLineLen];
  ostrstream logOss(logBuf, logLineLen);
  if (alogPrefix) logOss << alogPrefix;
  logOss << atitle << ": ";
  dumpRegValue(logOss, atype, abootpValue, 0);
  logOss << ends;
  logLine(mainWindow, logBuf);
  logOss.seekp(0);
  if (getValue(akeyName, atitle, (BYTE*)registryBuf, registryBufLen,
               atype, mainWindow)) {
    if (memcmp(registryBuf, (char*)abootpValue, registryBufLen)) {
      logOss << "...but registry has: ";
      dumpRegValue(logOss, atype, registryBuf, registryBufLen);
      logOss << ends;
      logLine(mainWindow, logBuf);
      logOss.seekp(0);
      discrepancy = TRUE;
    }
  }
  else {
    logOss << "...but registry has no value" << ends;
    logLine(mainWindow, logBuf);
    logOss.seekp(0);
    discrepancy = TRUE;
  }
  if (discrepancy &&
      (aforceUpdate ||
       (yesOrNo(mainWindow, "Fix this discrepancy?") == IDYES))) {
    HKEY keyHandle = getKey(akeyName, mainWindow);
    if (keyHandle == NULL) abortOnBadKey(akeyName, mainWindow);
    LONG rc = RegSetValueEx(keyHandle, atitle, 0, atype,
                            (const UCHAR*)abootpValue,
                            valueLen(abootpValue, atype));
    if (rc == ERROR_SUCCESS) {
      RegFlushKey(keyHandle);   // Do we need this in NT??
      logLine(mainWindow, "...fixed it!");
      result = 1;
    }
    else {
      char* msgBuf = new char[msgLen];
      ostrstream msgOss(msgBuf, msgLen);
      msgOss << "RegSetValueEx failed with error: " << rc << endl
             << "  for key: " << akeyName << endl
             << "  value name: " << atitle << ends;
      fail(mainWindow, msgBuf);
    }
    if (!isRootKey(keyHandle)) RegCloseKey(keyHandle);
  }
  return result;
}

/*
   Here we deal with a sequence of IP addresses in network byte order.
   The argument "anaddr" may point into the bootp receive buffer, either
   in the vendor area or the yiaddr field.  The "alen" argument is the number
   of bytes, not the number of IP addresses.  The optional "alimit" arg,
   which defaults to an essentially infinite value, limits the number
   of addresses to what the Registry value will accept without burping.
   (Thanks to Mikael Willberg for realizing that "NameServer" was such
   a value.)
*/

UINT WorkerParams::regIPAddrs(const char* akeyName, DWORD atype, char* atitle,
                              const UCHAR* anaddr, int alen, int alimit,
                              const char* alogPrefix)
{
  char* bootpBuf= new char[RegValueBufLen];
  ostrstream bootpOss(bootpBuf, RegValueBufLen);
  for (int entryNo = 1; alen > 0; entryNo++) {
    if (entryNo <= alimit) {
      bootpOss << WORD(anaddr[0]) << "."
        << WORD(anaddr[1]) << "."
        << WORD(anaddr[2]) << "."
        << WORD(anaddr[3]);
      if (atype == REG_MULTI_SZ) bootpOss << ends;
    }
    alen -= 4;
    if (alen > 0) {
      anaddr += 4;
      if ((entryNo < alimit) && (atype == REG_SZ)) {
        bootpOss << ((os == W95) ? "," : " ");
      }
    }
  }
  bootpOss << ends;
  return regify(akeyName, atype, atitle, (UCHAR*)bootpBuf, forceFix,
                alogPrefix);
}

/*
  Schedule a change of the computer's name for the next reboot if the
  new name supplied differs from the current one, returning a nonzero
  value if and only if a change has been scheduled.
*/

UINT WorkerParams::changeComputerName(char* anewName)
{
  char oldName[msgLen];
  DWORD nameLen = msgLen;
  GetComputerName(oldName, &nameLen);
  if (stricmp(oldName, anewName)) {
    SetComputerName(anewName);
    char* msgBuf = new char[msgLen];
    ostrstream msgOss(msgBuf, msgLen);
    msgOss << "Changed ComputerName from " << oldName << " to " << anewName
           << ends;
    logLine(mainWindow, msgBuf);
    return 1;
  }
  else return 0;
}

/*
   RFC 1084's hostname may include the domain suffix; if so we need to
   register the latter separately.
*/

UINT WorkerParams::regHostname(RFC1084Field* f)
{
#if 0
  char* computerKeyName = "HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\"
    "Control\\ComputerName\\ComputerName";
#endif
  UINT result = 0;
  char* bootpBuf = new char[f->length + 1];
  ostrstream bootpOss(bootpBuf, f->length + 1);
  UCHAR* firstDot = (UCHAR*)memchr(f->data, '.', f->length);
  if (firstDot == NULL) { // No domain
    for (UCHAR i = 0; i < f->length; i++) bootpOss << f->data[i];
    bootpOss << ends;
    if (!ignoreHostName) {
      result += regify(tcpKeyName, REG_SZ, "HostName", (UCHAR*)bootpBuf,
                       forceFix);
    }
    if (updateComputerName) {
#if 0
      result += regify(computerKeyName, REG_SZ, "ComputerName",
                       (UCHAR*)bootpBuf, forceFix);
#else
      result += changeComputerName(bootpBuf);
#endif
    }
  }
  else {
    size_t preDomainLength = firstDot - f->data;
    for (size_t i = 0; i < preDomainLength; i++) bootpOss << f->data[i];
    bootpOss << ends;
    result += regify(tcpKeyName, REG_SZ, "HostName", (UCHAR*)bootpBuf,
                    forceFix);
    if (updateComputerName) {
#if 0
      result += regify(computerKeyName, REG_SZ, "ComputerName",
                       (UCHAR*)bootpBuf, forceFix);
#else
      result += changeComputerName(bootpBuf);
#endif
    }
    bootpOss.seekp(0); // Rewind
    for (UCHAR j =  preDomainLength + 1; j < f->length; j++) {
      bootpOss << f->data[j];
    }
    bootpOss << ends;
    result += regify(tcpKeyName, REG_SZ, "Domain", (UCHAR*)bootpBuf, forceFix);
  }
  return result;
}

/*
   RFC 1533's domain name:
*/

UINT WorkerParams::regDomainName(RFC1084Field* f)
{
  char* bootpBuf = new char[f->length + 1];
  ostrstream bootpOss(bootpBuf, f->length + 1);
  for (UCHAR i = 0; i < f->length; i++) bootpOss << f->data[i];
  bootpOss << ends;
  return regify(tcpKeyName, REG_SZ, "Domain", (UCHAR*)bootpBuf, forceFix);
}

/*
  RFC 1533's NetBIOS node type, which comes to us as a binary byte of 1, 2,
  4 or 8:
*/

UINT WorkerParams::regNetBIOSNodeType(RFC1084Field* f)
{
  char* logPrefix = "NetBIOS ";
  if (os == WNT) {
    char* netBTParamsKeyName =
      "HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\"
      "Services\\NetBT\\Parameters";
    DWORD data = f->data[0];
    return regify(netBTParamsKeyName, REG_DWORD_LITTLE_ENDIAN, "NodeType",
                  &data, forceFix, logPrefix);
  }
  else {
    char data[10];
    wsprintf(data, "%d", f->data[0]);
    return regify(tcpKeyName, REG_SZ, "NodeType", data, forceFix, logPrefix);
  }
}

/*
   RFC 1533's NetBIOS name server = WINS server:
*/

UINT WorkerParams::regNetBIOSNameServer(RFC1084Field* f)
{
  UINT restart = 0;
  if (os == WNT) {
    char* netBTAdaptersPrefix =
      "HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\"
      "Services\\NetBT\\Adapters\\";
    char* netcardNetBTKeyName =
      new char[strlen(netBTAdaptersPrefix) + strlen(adapterName) + 1];
    strcpy(netcardNetBTKeyName, netBTAdaptersPrefix);
    strcat(netcardNetBTKeyName, adapterName);
    restart +=
      regIPAddrs(netcardNetBTKeyName, REG_SZ, "NameServer", f->data, 4, 4,
                 "NetBIOS ");
    if (f->length > 4) {
      restart += regIPAddrs(netcardNetBTKeyName, REG_SZ, "NameServerBackup",
                            &f->data[4], 4, 4, "NetBIOS ");
    }
    delete netcardNetBTKeyName;
  }
  else {
    restart += regIPAddrs(tcpKeyName, REG_SZ, "NameServer1", f->data, 4, 4,
                          "NetBIOS ");
    if (f->length > 4) {
      restart += regIPAddrs(tcpKeyName, REG_SZ, "NameServer2", &f->data[4], 4,
                            4, "NetBIOS ");
    }
  }
  return restart;
}

/*
  This function returns the last segment of a Registry key, starting with
  the character after the backslash, if there is a backslash.
*/

char* keyNameTail(char* akeyName)
{
  char* p;
  for (p = akeyName + strlen(akeyName) - 1; *p != '\\'; p--) {
    if (p == akeyName) return p; // no backslash; simple keyname
  }
  return (p + 1);
}

/*
  Windows 95: Is this a dial-up or virtual private networking adapter?

  As far as we're concerned instances of TCP/IP over connections
  other than direct-to-netcard don't count.

  Our search argument will be of the form HKLM\Enum\Network\MSTCP\000x.
  We want to see if there's a non-netcard HKLM\Enum\Root\Net\000y bound to it.
*/

BOOL notThroughNetcard(char* amstcpKeyName, HWND amainWindow)
{
  char* rootNetParentKeyName = "HKEY_LOCAL_MACHINE\\Enum\\Root\\Net";
  BOOL result = FALSE;
  HKEY rootNetParentHandle = getKey(rootNetParentKeyName, amainWindow, TRUE);
  if (rootNetParentHandle != NULL) { // There might be a dial-up "adapter"
    for (DWORD n = 0; ; n++) {
      char rootNetSubKeyName[80]; // HKLM\Enum\Root\Net\000y tail
      DWORD rootNetSubKeyNameSize = sizeof(rootNetSubKeyName);
      FILETIME lastUpdate;

      LONG rc = RegEnumKeyEx(rootNetParentHandle, n, rootNetSubKeyName,
                             &rootNetSubKeyNameSize, NULL, NULL, NULL,
                             &lastUpdate);
      if (rc != ERROR_SUCCESS) break; // Haven't found dial-up adapter binding.

      /* We have an adapter; is it Dial-up? */
      char rootNetKeyName[100];
      strcpy(rootNetKeyName, rootNetParentKeyName);
      strcat(rootNetKeyName, "\\");
      strcat(rootNetKeyName, rootNetSubKeyName);
      char deviceDesc[100];       // e.g. "Dial-up Adapter"
      DWORD deviceDescLen = sizeof(deviceDesc);
      if (getValue(rootNetKeyName, "DeviceDesc", (BYTE*)deviceDesc,
                   deviceDescLen, REG_SZ, amainWindow)) {
        char* dialupSignature = "Dial";
        char* vpnaSignature = "Microsoft Virtual Private";
        if (!strnicmp(deviceDesc, dialupSignature, strlen(dialupSignature)) ||
            !strnicmp(deviceDesc, vpnaSignature, strlen(vpnaSignature))) {
          char bindingValue[20];       // This will be empty
          DWORD bindingValueLen = sizeof(bindingValue);
          char bindingName[40]; // e.g. "MSTCP\0001"
          strcpy(bindingName, "MSTCP\\");
          strcat(bindingName, keyNameTail(amstcpKeyName));
          /* Append "Bindings" to dial-up adapter key name */
          strcat(rootNetKeyName, "\\Bindings"); 
          if (getValue(rootNetKeyName, bindingName, (BYTE*)bindingValue,
                       bindingValueLen, REG_SZ, amainWindow)) {
            result = TRUE;
            break;
          }
        }
      }
    }
    RegCloseKey(rootNetParentHandle);
  }
  return result;
}

/*
  Windows 95 TCP Registry navigation:

  Here we find the subkey under netTransKeyName containing TCP/IP info
  that uses a LAN adapter, i.e. that isn't bound to an instance of
  Dial-up Networking.  We insist that it be unique.

  The string we return is the caller's responsibility to delete.
  
  To find our key we don't traverse the ...NetTrans keys directly; not
  all of them will be for TCP/IP and, besides, not all of them need be
  active.  Instead, we traverse the keys under HKLM\Enum\Network\MSTCP.
  For each of these we check if it's a binding for a dial-up "adapter";
  if not, we remember which NetTrans subkey (pointed to by the "Driver"
  value) it's associated with.
*/

char* WorkerParams::findTCPNetTransKeyName()
{
  char* mstcpParentKeyName = "HKEY_LOCAL_MACHINE\\Enum\\Network\\MSTCP";
  HKEY ntParentHandle = getKey(netTransKeyName, mainWindow);
  if (ntParentHandle == NULL) abortOnBadKey(netTransKeyName, mainWindow);
  HKEY mstcpParentHandle = getKey(mstcpParentKeyName, mainWindow);
  if (mstcpParentHandle == NULL) abortOnBadKey(mstcpParentKeyName, mainWindow);
  size_t nTCPKeys = 0;          // Want to end up with 1
  char tcpntKeyNameTail[80] = { 0 }; // e.g. "0003"

  for (DWORD n = 0; ; n++) {
    char mstcpSubKeyName[80];   // e.g. "0001"
    DWORD mstcpSubKeyNameSize = sizeof(mstcpSubKeyName);
    FILETIME lastUpdate;

    LONG rc = RegEnumKeyEx(mstcpParentHandle, n, mstcpSubKeyName,
                           &mstcpSubKeyNameSize, NULL, NULL, NULL,
                           &lastUpdate);
    if (rc != ERROR_SUCCESS) break; // No more subkeys.

    char mstcpKeyName[120];
    strcpy(mstcpKeyName, mstcpParentKeyName);
    strcat(mstcpKeyName, "\\");
    strcat(mstcpKeyName, mstcpSubKeyName);

    if (notThroughNetcard(mstcpKeyName, mainWindow)) continue; // Uninterested!
    else if (++nTCPKeys > 1) break; // Bail out if more than one!

    /* We have an instance of MSTCP worth keeping; save its NetTrans tail */
    char driverName[100];       // e.g. "NetTrans\0003"
    DWORD driverNameLen = sizeof(driverName);
    if (getValue(mstcpKeyName, "Driver", (BYTE*)driverName, driverNameLen,
                 REG_SZ, mainWindow)) {
      strcpy(tcpntKeyNameTail, keyNameTail(driverName));
    }
    else fail(mainWindow, "Illogical MSTCP enumerator key");
  }

  RegCloseKey(ntParentHandle);
  RegCloseKey(mstcpParentHandle);

  if (nTCPKeys > 1) fail(mainWindow,
                         "No unique Registry key for TCP/IP over LAN");
  else if (nTCPKeys == 0) fail(mainWindow,
                               "Can't find Registry key for LAN TCP/IP");

  /* Compute and return the key */
  char* result = new char[strlen(netTransKeyName) +
                         strlen(tcpntKeyNameTail) + 4];
  strcpy(result, netTransKeyName);
  strcat(result, "\\");
  strcat(result, tcpntKeyNameTail);
  return result;
}

/*
  Starting without WinSock initialized, make ourselves a socket with which we
  can broadcast bootp requests and receive bootp replies.  Due to the way NT,
  at least, sometimes starts up raggedly, we may want to give it another chance
  or so to bind after a decent interval.  If, however, bind() fails
  persistently with WSAEADDRNOTAVAIL, we fail silently on the assumptions
  that the netcard isn't available and that the user wants it that way.
*/

SOCKET WorkerParams::getBootpSocket()
{
  WSADATA wskData;
  char* msgBuf = NULL;
  SOCKET sock = INVALID_SOCKET;
  BOOL allowBroadcast = TRUE;
  int bindError;
  if (WSAStartup(MAKEWORD(1, 1), &wskData)) {
    msgBuf = "Can't find Winsock";
  }
  else if ((sock = socket(PF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET) {
    msgBuf = "Unable to create a socket";
  }
  else if (setsockopt(sock, SOL_SOCKET, SO_BROADCAST,
                      (const char*)&allowBroadcast,
                      sizeof(allowBroadcast)) != 0) {
    msgBuf = "Unable to allow broadcast";
  }
  else {
    sockaddr_in ourAddr;
    char ipAddrBuf[80];
    DWORD ipAddrBufLen = sizeof(ipAddrBuf);
    memset (&ourAddr, 0, sizeof(ourAddr));
    ourAddr.sin_family = AF_INET;
    if (!getValue(transportKeyName, "IPAddress", (BYTE*)ipAddrBuf,
                  ipAddrBufLen, transportValueType, mainWindow)) {
      fail(mainWindow, "Can't find our IP address");
    }
    ourAddr.sin_addr.s_addr = inet_addr(ipAddrBuf);
    if (ourAddr.sin_addr.s_addr == INADDR_NONE) fail(mainWindow,
                                                     "Invalid IP address");
    ourAddr.sin_port = htons(PortBootpC);
    int bindResult, bindRetries = 20, bindWaitSecs = 15;
    while ((bindResult = bind(sock, (sockaddr*)&ourAddr, sizeof(sockaddr_in)))
           != 0) {
      if (--bindRetries <= 0) break;
      bindError = WSAGetLastError();
      if (bindError != WSAEADDRINUSE) break;
      Sleep(bindWaitSecs * 1000);
    }
    if (bindResult != 0) {
      msgBuf = new char[logLineLen];
      ostrstream msgOss(msgBuf, logLineLen);
      msgOss << "Unable to bind to bootp client port: "
             << bindError << ends;
    }
  }
  if (msgBuf) fail(mainWindow, msgBuf, BOOL(bindError == WSAEADDRNOTAVAIL));
  return sock;
}

/*
  Here (under NT) we try to adjust our standing in the world so we qualify
  to reboot Windows, returning TRUE if we succeed.  Thanks to Gary Dobbins
  for this code!
*/

BOOL WorkerParams::grabRebootPrivilege()
{
  HANDLE hToken;
  TOKEN_PRIVILEGES tkp;

  // Need to change privileges
  if(!OpenProcessToken(GetCurrentProcess(),
                       TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken)) {
    logLastError(mainWindow, "OpenProcessToken");
    return FALSE;
  }

  // Get the LUID for shutdown privilege
  if (!LookupPrivilegeValue(NULL, TEXT("SeShutdownPrivilege"),
                            &tkp.Privileges[0].Luid)) {
    logLastError(mainWindow, "LookupPrivilegeValue");
    return FALSE;
  }

  tkp.PrivilegeCount = 1;  // one privilege to set
  tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

  // Get shutdown privilege for this process.
  if(!AdjustTokenPrivileges(hToken, FALSE, &tkp, 0,
                            (PTOKEN_PRIVILEGES)NULL, 0)) {
    logLastError(mainWindow, "AdjustTokenPrivileges");
    return FALSE;
  }
  logLine(mainWindow, "Privilege was granted");
  return TRUE;
}

/*
   Here we log everything we wrote to the listbox.
*/

void writeLog(HWND hDlg)
{
  enum { lineBufLen = 500 };
  char* lineBuf = new char[lineBufLen];
  LRESULT lineLen;
  for (USHORT lineNo = 0;
       (lineLen = SendDlgItemMessage(hDlg, IDC_LOGBOX, LB_GETTEXTLEN,
                                     lineNo, 0)) != LB_ERR;
       lineNo++) {
    if (lineLen < lineBufLen) {
      SendDlgItemMessage(hDlg, IDC_LOGBOX, LB_GETTEXT, lineNo,
                         (LPARAM)lineBuf);
      lineBuf[lineLen] = '\0';
      outFile << lineBuf << endl;
    }
    else outFile << "**Listbox line too long: " << lineLen << endl;
  }
  delete lineBuf;
}

/*
  Why do we distinguish between NT and 95 here?  Because ExitWindowsEx()
  doesn't work when invoked by a service-launched billgPC at system boot
  time, that's why.  We'd get a TRUE return value but the system wouldn't
  reboot and NT would be confused in an interesting way: if a user logged on
  and then invoked Task Manager, the system would start to shut down and
  then just hang.  (This was NTWS 4.0)

  Another annoying detail here is the need to write the log and quiesce our
  I/O connections before asking the OS to reboot.  This is because the OS
  seems not to want to close our main window, which would normally take
  care of these needs, as part of the reboot.
*/

BOOL WorkerParams::doReboot()
{
#if 1                           // Won't do it in automatic service?
  UINT rebootFlags = EWX_REBOOT;
#else  // crashes Win95!
  UINT rebootFlags = EWX_REBOOT | EWX_LOGOFF | EWX_FORCE;
#endif
  writeLog(mainWindow);
  shutdown();                   // Do all we need to stabilize ourselves.
  if (os == WNT) return InitiateSystemShutdown(NULL, "billgPC is rebooting...",
                                               15, // time to display message
                                               FALSE, // force unsaved apps?
                                               TRUE); // reboot?
  else return ExitWindowsEx(rebootFlags, 0);
}

void WorkerParams::tryReboot()
{
#if 0
  if (os == WNT) {
    logLine(mainWindow, "Waiting for system stability before we reboot...");
    Sleep(60000);
  }
#endif
  logLine(mainWindow, "Trying to reboot");
  if (!doReboot()) {
    logLine(mainWindow, "Grabbing privilege");
    if (!grabRebootPrivilege() || !doReboot()) {
      if (os == WNT) logLastError(mainWindow, "Rebooting");
      else alert(mainWindow, "Windows won't reboot - please do it manually");
    }
    else logLine(mainWindow, "Trying to reboot with privilege");
  }
}

/*
   Here we give the user the chance to reboot the computer.  Note that if
   forceReboot is nonzero yesOrNo doesn't get evaluated.
*/

void WorkerParams::allowReboot()
{
  if (forceReboot ||
      yesOrNo(mainWindow,
              "Your computer's settings have been changed to operate\n"
              "on this network.\n"
              "System restart is needed for settings to take effect.\n"
              "Proceed with restart?") == IDYES) tryReboot();
}

void WorkerParams::logServerAddress(UCHAR* anaddr)
{
  char* bootpBuf= new char[RegValueBufLen];
  ostrstream bootpOss(bootpBuf, RegValueBufLen);
  bootpOss << "Bootp server address: "
           << WORD(anaddr[0]) << "."
           << WORD(anaddr[1]) << "."
           << WORD(anaddr[2]) << "."
           << WORD(anaddr[3]) << ends;
  logLine(mainWindow, bootpBuf);
}

/*
   The function that processes the bootp response against the Registry
   returns a nonzero value if there's a reason to reboot.  Note that we may
   have to translate the RFC 1084 tag into another tag if we were instructed
   to do so by the command line.
*/

UINT WorkerParams::registerResponse(BootpBuf& buf)
{
  logServerAddress((UCHAR*)&buf.siaddr);
  UINT restart = regIPAddrs(transportKeyName, transportValueType,  "IPAddress",
                            ((UCHAR*)&buf.yiaddr), 4);
  RFC1084Field* f;
  for (RFC1084Iterator iter(buf); (f = iter.more()) != NULL; iter.next()) {
    RFC1084Tag tag = f->tagXlateTbl[f->tag] ?
      RFC1084Tag(f->tagXlateTbl[f->tag]) : RFC1084Tag(f->tag);
    switch(tag)
    {
    case SubnetMask:
      restart += regIPAddrs(transportKeyName, transportValueType,
                            subnetMaskString, f->data, f->length);
      break;
    case Gateways:
      restart += regIPAddrs(transportKeyName, transportValueType,
                            "DefaultGateway", f->data, f->length);
      break;
    case DomainNameServers:
      enum { MaxNameServers = 3 }; // That's all the Registry will take.
      restart += regIPAddrs(tcpKeyName, REG_SZ, "NameServer", f->data,
                            f->length, MaxNameServers, "DNS ");
      if (os == W95) {
        restart += regify(tcpKeyName, REG_SZ, "EnableDNS", "1", forceFix);
      }
      break;
    case Hostname:
      restart += regHostname(f);
      break;
    case DomainName:
      restart += regDomainName(f);
      break;
    case NetBIOSNameServer:
      restart += regNetBIOSNameServer(f);
      break;
    case NetBIOSNodeType:
      restart += regNetBIOSNodeType(f);
      break;
    default:
      {
        if (f->tag > MaxOptionSeen) {
          char* msgBuf = new char[msgLen];
          ostrstream msgOss(msgBuf, msgLen);
          msgOss << "Unhandled RFC 1084/1533 code: " << WORD(f->tag) << ends;
          alert(mainWindow, msgBuf);
        }
        break;
      }
    }
  }
  return restart;
}

void WorkerParams::openNTBgps()
{
  enum { SYSVersionMin = 129, SYSVersionMax = 131 };
  if (srvHandle != NULL) {      // Service exists
    char bgpsDeviceName[80];
    wsprintf(bgpsDeviceName, "\\\\.\\billgPS%s", adapterName);
    ntBgpsHandle = CreateFile(bgpsDeviceName,
                              GENERIC_READ | GENERIC_WRITE, 0, NULL,
                              OPEN_EXISTING,
                              FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED,
                              NULL);
    if (ntBgpsHandle != INVALID_HANDLE_VALUE) {
      logLine(mainWindow, "Opened billgPS device");
      UINT sysVersion;
      DWORD bytesReturned;
      if (DeviceIoControl(ntBgpsHandle, DIOC_BillgpGetVersion,
                          NULL, 0, &sysVersion, sizeof(sysVersion), 
                          &bytesReturned, NULL)) {
        if ((sysVersion < SYSVersionMin) || (sysVersion > SYSVersionMax)) {
          logLine(mainWindow, "Wrong version of billgPS.SYS");
          CloseHandle(ntBgpsHandle);
          ntBgpsHandle = INVALID_HANDLE_VALUE;
        }
        else {
          char* msgBuf = new char[logLineLen];
          ostrstream msgOss(msgBuf, logLineLen);
          msgOss << "Using billgPS.SYS version " <<
            (sysVersion / 100) << "." << setw(2) << setfill('0') <<
            (sysVersion % 100) << ends;
          logLine(mainWindow, msgBuf);
        }
      }
      else {
        logLine(mainWindow, "Can't get version of billgPS.SYS");
        CloseHandle(ntBgpsHandle);
        ntBgpsHandle = INVALID_HANDLE_VALUE;
      }
    }
    else logLastError(mainWindow, "Opening billgPS device");
  }
}

/*
  Windows NT TCP Registry navigation, pre-NT 5:

  The "Bind" value of the Tcpip Linkage key has a series of strings of the
  form \Device\adaptername, where if adaptername begins with "NdisWan" it's
  a Dial-up pseudo-adapter.  We want one entry to remain.  If so, that
  adapter's Parameters\Tcpip key name is what we return.

  While we're at it we also allocate and fill adapterName so we can find
  the MAC address later.

  The string we return is the caller's responsibility to delete.
*/

char* WorkerParams::findNT4NetcardTCPParams()
{
  char* tcpLinkageKeyName = "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\"
    "Services\\Tcpip\\Linkage";
  char* srvKeyPrefix = "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\"
    "Services\\";
  char* tcpParamSuffix = "\\Parameters\\Tcpip";
  DWORD bindBufLen = 1000;
  char* bindBuf = new char[bindBufLen];
  if (!getValue(tcpLinkageKeyName, "Bind", (BYTE*)bindBuf, bindBufLen,
                REG_MULTI_SZ, mainWindow)){
    fail(mainWindow, "Registry failure");
  }
  size_t nTCPKeys = 0;          // Want to end up with 1
  for (CHAR* thisValue = bindBuf;
       *thisValue;
       thisValue += (strlen(thisValue) + 1)) {
    if (strstr(thisValue, "NdisWan")) continue; // Not interested in Dial-up!
    else if (++nTCPKeys > 1) break; // Bail out if more than one!
    adapterName = new char[strlen(keyNameTail(thisValue)) + 1];
    strcpy(adapterName, keyNameTail(thisValue));
  }
  if (nTCPKeys > 1) fail(mainWindow,
                         "No unique Registry key for TCP/IP over LAN");
  else if (nTCPKeys == 0) fail(mainWindow,
                               "Can't find Registry key for LAN TCP/IP");

  /* Compute and return the key */
  char* result = new char[strlen(srvKeyPrefix) + strlen(adapterName) +
                         strlen(tcpParamSuffix) + 1];
  strcpy(result, srvKeyPrefix);
  strcat(result, adapterName);
  strcat(result, tcpParamSuffix);
  delete bindBuf;
  openNTBgps();
  return result;
}

/*
  Windows NT TCP Registry navigation, NT 5:

  We're interested in the network adapters enumerated below the key for
  the network adapter class, which in NT 5 involves a GUID.  For each one,
  we're only interested if it's bound above to TCP/IP and below to Ethernet
  or Token Ring, and we're only interested if there's one and only one
  meeting our criteria.  Then computing the keyname is easy after we open
  the corresponding billgPS device.
*/

char* WorkerParams::findNT5NetcardTCPParams()
{
  char* netcardClassKeyName =
    "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Control\\Class\\"
    "{4D36E972-E325-11CE-BFC1-08002BE10318}";
  char* tcpParamsPrefix =
    "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\Tcpip\\"
    "Parameters\\Interfaces\\";
  enum { KeyNameBufLen = 1000, ValueBufLen = 1000 };
  DWORD valueBufLen;
  char* valueBuf = new char[ValueBufLen];
  char* keyNameBuf = new char[KeyNameBufLen];
  size_t nTCPKeys = 0;          // Want to end up with 1
  HKEY netcardClassHandle = getKey(netcardClassKeyName, mainWindow);
  if (netcardClassHandle == NULL) {
    abortOnBadKey(netcardClassKeyName, mainWindow);
  }

  for (DWORD adapterNo = 0; ; adapterNo++) {
    char netcardSubKeyName[80];   // e.g. "0001"
    DWORD netcardSubKeyNameSize = sizeof(netcardSubKeyName);
    FILETIME lastUpdate;

    LONG rc = RegEnumKeyEx(netcardClassHandle, adapterNo, netcardSubKeyName,
                           &netcardSubKeyNameSize, NULL, NULL, NULL,
                           &lastUpdate);
    if (rc != ERROR_SUCCESS) break; // No more subkeys.

    strcpy(keyNameBuf, netcardClassKeyName);
    strcat(keyNameBuf, "\\");
    strcat(keyNameBuf, netcardSubKeyName); // e.g. "0000"
    size_t cardKeyNameLen = strlen(keyNameBuf); // length so far...
    strcat(keyNameBuf, "\\Linkage");
    valueBufLen = ValueBufLen;
    if (getValue(keyNameBuf, "UpperBind", (BYTE*)valueBuf, valueBufLen,
                 REG_MULTI_SZ, mainWindow)) {
      CHAR* upperBindP;
      for (upperBindP = valueBuf;
           *upperBindP && stricmp(upperBindP, "Tcpip");
           upperBindP += (strlen(upperBindP) + 1)) { }
      if (*upperBindP == 0) continue; // Adapter isn't bound to TCP/IP.
      // NT5 adapter key has no Parameters subkey w/MediaType
      valueBufLen = ValueBufLen;
      strcpy(keyNameBuf + cardKeyNameLen, "\\Ndi\\Interfaces");
      if (getValue(keyNameBuf, "DefLower", (BYTE*)valueBuf, valueBufLen,
                   REG_SZ, mainWindow)&&
          ((!stricmp(valueBuf, "ethernet")) ||
           (!stricmp(valueBuf, "token ring")))) {
        valueBufLen = ValueBufLen;
        strcpy(keyNameBuf + cardKeyNameLen, "\\Linkage");
        if (getValue(keyNameBuf, "Bind", (BYTE*)valueBuf, valueBufLen,
                     REG_MULTI_SZ, mainWindow)) {
          if (++nTCPKeys > 1) break;
          adapterName = new char[strlen(keyNameTail(valueBuf)) + 1];
          strcpy(adapterName, keyNameTail(valueBuf));
        }
      }
    }
  }

  RegCloseKey(netcardClassHandle);

  if (nTCPKeys > 1) fail(mainWindow,
                         "No unique Registry key for TCP/IP over LAN");
  else if (nTCPKeys == 0) fail(mainWindow,
                               "Can't find Registry key for LAN TCP/IP");

  /* Compute and return the key */
  char* result = new char[strlen(tcpParamsPrefix) + strlen(adapterName) + 1];
  strcpy(result, tcpParamsPrefix);
  strcat(result, adapterName);
  openNTBgps();
  delete valueBuf;
  delete keyNameBuf;
  return result;
}

BOOL WorkerParams::isDHCP()
{
  BOOL result = FALSE;
  char valueBuf[80];
  DWORD valueBufLen = sizeof(valueBuf);
  if (getValue(transportKeyName, "IPAddress", (BYTE*)valueBuf, valueBufLen,
               transportValueType, mainWindow)) {
    if (strlen(valueBuf) == 0) result = TRUE;
    else if (!strcmp(valueBuf, "0.0.0.0")) result = TRUE;
  }
  else result = TRUE;
  return result;
}

/*
  This function forces the IP address to a pseudo-random address based on
  the last 3 bytes of the MAC address, which makes it *extremely* unlikely
  that two machines on the same subnet will yield the same values.  It returns
  1 if it actually changed the IP address (i.e. if the address wasn't the
  pseudo-random value already) and 0 if it didn't.
*/

UINT WorkerParams::forcePseudoRandomIPAddress(HANDLE avxdHandle)
{
  DWORD hwAddrSize = MacAddressSize;
  BYTE hwAddrBuf[MacAddressSize];
  BYTE hwType;
  logLine(mainWindow, "Computing pseudo-random IP address...");
  getHardwareAddress(hwAddrBuf, hwAddrSize, hwType, avxdHandle, FALSE);
  char* msgBuf = new char[msgLen];
  ostrstream msgOss(msgBuf, msgLen);
  msgOss << "1";
  for (size_t i = MacAddressSize - 3; i < MacAddressSize; i++) {
    msgOss << "." << (hwAddrBuf[i] ? WORD(hwAddrBuf[i]) : 1);
  }
  msgOss << ends << ends;
  return regify(transportKeyName, transportValueType, "IPAddress", msgBuf,
                TRUE);
}

/*
  The IP address and subnet mask strings have extra trailing null characters
  so they can function as REG_MULTI_SZ under NT.  The IP address is randomized
  so many machines can run billgPC in parallel in an installation where all
  machines' disk images are downloaded uniformly.
*/

void WorkerParams::setupBootp(HANDLE avxdHandle)
{
  logLine(mainWindow, "Setting up BOOTP");
  logLine(mainWindow, "Don't worry: some settings are provisional");
  logLine(mainWindow, "...and will be changed the next time billgPC runs.");
  forcePseudoRandomIPAddress(avxdHandle);
  regify(transportKeyName, transportValueType, subnetMaskString,
         "255.255.255.0\0", TRUE);
  if (os == W95) {
    regify(tcpKeyName, REG_SZ, "Lanabase", "0", TRUE);
    regify(tcpKeyName, REG_SZ, "NodeType", "1", TRUE);
  }
  else{
    DWORD no = 0;
    regify(transportKeyName, REG_DWORD_LITTLE_ENDIAN, "EnableDHCP", &no, TRUE);
  }
}

/*
  This applies only to Win95.
*/

BOOL WorkerParams::stackIsMicrosoft()
{
  char stackNameBuf[80];
  DWORD stackNameSize = sizeof(stackNameBuf);
  char ndiTail[] = "\\Ndi";
  char* transNdiKeyName = new char[strlen(transportKeyName) +
                                  strlen(ndiTail) + 1]; // Thanks, Uwe Proemm!
  strcpy(transNdiKeyName, transportKeyName);
  strcat(transNdiKeyName, ndiTail);
  BOOL rc = getValue(transNdiKeyName, "DeviceID",
                     (BYTE*)stackNameBuf, stackNameSize, REG_SZ,
                     mainWindow);
  delete transNdiKeyName;
  if (strcmp(stackNameBuf, "MSTCP")) {
    char* msgBuf = new char[msgLen];
    ostrstream msgOss(msgBuf, msgLen);
    msgOss << "This program requires the Microsoft TCP/IP stack.";
    if (rc) msgOss << endl << "This machine appears to be running "
      << stackNameBuf << ".";
    msgOss << ends;
    fail(mainWindow, msgBuf);
    return FALSE;
  }
  return TRUE;
}

/*
   Here we check if the machine we're running on really uses the Microsoft
   IP stack.  We also make sure it isn't set up for DHCP, resorting to
   setting up the Registry for bootp and rebooting should that be needed.
*/

void WorkerParams::checkStack(HANDLE avxdHandle)
{
  transportKeyName = (os == W95) ? findTCPNetTransKeyName() :
    ((osMajorVersion < 5) ? findNT4NetcardTCPParams() :
     findNT5NetcardTCPParams());
  if ((os == W95) && !stackIsMicrosoft()) _endthread();
  else if (isDHCP()) {
    if ((forceFix && forceReboot) ||
        yesOrNo(mainWindow, "This computer is set up for DHCP not BOOTP.\n"
                "For billgPC to succeed we first need to set up BOOTP "
                "and then restart Windows.\n"
                "Then you can rerun billgPC and it should work.\n"
                "Do you want to set up BOOTP and restart Windows?") == IDYES) {
      setupBootp(avxdHandle);
      tryReboot();
      SendMessage(mainWindow, WM_CLOSE, 0, 0);
    }
    else fail(mainWindow, "Sorry - billgPC is incompatible with DHCP.");
  }
}

/*
  Diagnostic dump of the VxD (obviously only for 95)
*/

void WorkerParams::dumpDriverMemory(HANDLE avxdHandle)
{
  if (os == WNT) {
    if (ntBgpsHandle == INVALID_HANDLE_VALUE) {
      fail(mainWindow, "Driver dump requires billgPS.SYS");
    }
    else {
      enum { BufSize = 2000 };
      BYTE* buf = new BYTE[BufSize];
      DWORD bytesReturned;
      if (DeviceIoControl(ntBgpsHandle, DIOC_BillgpDumpDriver,
                          NULL, 0, buf, BufSize, 
                          &bytesReturned, NULL)) {
        dump(mainWindow, bytesReturned, buf, "Driver context:");
      }
      else logLine(mainWindow, "Can't dump billgPS.SYS");
      delete buf;
    }
  }
  else if (avxdHandle == INVALID_HANDLE_VALUE) {
    fail(mainWindow, "Driver dump requires billgps.vxd");
  }
  else {
    enum { BufSize = 2000 };
    BYTE* buf = new BYTE[BufSize];
    DWORD bytesReturned;
    if (!DeviceIoControl(avxdHandle, DIOC_BillgpDumpDriver,
                         NULL, 0, buf, BufSize, 
                         &bytesReturned, NULL)) {
      fail(mainWindow, "Can't do DIOC_BillgpDumpDriver IOCTL");
    }
    else {
      dump(mainWindow, bytesReturned, buf, "VxD context:");
      delete buf;
      PostMessage(mainWindow, WM_CLOSE, 0, 0);
    }
  }
  _endthread();
}

/*
  Because billgPC may be running as part of a startup procedure on a machine
  (e.g. a laptop) for which it's normal to be disconnected from any LAN, we
  need a way to tell if we're disconnected so we won't complain when we
  find that the receiver thread never got anything.

  We assert we're disconnected unless our first gateway rejects our illegal
  attempt to connect to a bootpd on it via TCP (not UDP!)  Because Win95
  uses REG_SZ, not REG_MULTI_SZ, for the gateway addresses, we may need to
  change a command to a null character before calling inet_addr().
*/

BOOL WorkerParams::disconnected()
{
#if 1                           // We don't trust this!
  return FALSE;
#else
  BOOL result = TRUE;
  char gatewayAddrs[RegValueBufLen];
  DWORD registryBufLen = RegValueBufLen;
  *gatewayAddrs = '\0';         // Just in case...
  if (getValue(transportKeyName, "DefaultGateway", (BYTE*)gatewayAddrs,
               registryBufLen, transportValueType, mainWindow) &&
      (strlen(gatewayAddrs) > 0)) {
    sockaddr_in theirAddr;
    memset(&theirAddr, 0, sizeof(theirAddr));
    theirAddr.sin_family = AF_INET;
    theirAddr.sin_port = PortBootpS;
    char* firstComma = strchr(gatewayAddrs, ',');
    if (firstComma != NULL) *firstComma = '\0';
    theirAddr.sin_addr.s_addr = inet_addr(gatewayAddrs);
    SOCKET sock = socket(PF_INET, SOCK_STREAM, 0);
    if (sock != INVALID_SOCKET) {
      if (connect(sock, (sockaddr*)&theirAddr, sizeof(theirAddr)) < 0) {
        if (WSAGetLastError() == WSAECONNREFUSED) result = FALSE;
      }
      quiesceSocket(sock, FALSE);
    }
  }
  return result;
#endif
}

/* In typical cases of bootp failure we just put up a message box.  If
  we're authorized to force Registry fixes and reboot without user
  intervention, though, we try to "randomize" the IP address and
  reboot unless the IP address already had the pseudo-random value
  corresponding to the machine's MAC address.  The reason for this is
  that the Registry moves with the hard disk, not the netcard; in an
  installation where hard disks are removable there may be IP address
  conflicts that would defeat the operation of billgPC otherwise.
*/

void WorkerParams::handleBootpFailure(HANDLE avxdHandle)
{
  ShowWindow(mainWindow, SW_RESTORE); // ***temp***
  if (forceFix && forceReboot && // We have the right to randomize & reboot
      (srvHandle == NULL) &&    // NT billgPS obviates randomization
      forcePseudoRandomIPAddress(avxdHandle)) { // and IP wasn't randomized.
    tryReboot();
    SendMessage(mainWindow, WM_CLOSE, 0, 0);
  }
  else ourMessageBox(mainWindow,
                     "No BOOTP configuration is available from this network."
                     "\nIf your computer is not connected to a network,\n"
                     "this is not an error.\n"
                     "Otherwise contact your network administrator for help.",
                     MB_ICONEXCLAMATION);
}

enum { ThreadStackSize = 8192 };

/*
   We use _beginthread() and _endthread() rather than the nicer CreateThread()
   etc. because Microsoft says using the latter in a program perverse enough
   to use the C runtime library causes a memory leak.  Whether or not that's
   a bug, to ignore the hint would be to punish Microsoft for its
   candor.  Wouldn't that be the wrong thing to punish them for?
*/

void worker(void* aparams)
{
  WorkerParams& params = *((WorkerParams*)aparams);
  BootpBuf sendBuf;
  BigBootpBuf recvBuf;
  HANDLE vxdHandle = getBillgpsHandle(params.mainWindow);
  if ((os == W95) && params.dumpDriver) params.dumpDriverMemory(vxdHandle);
  params.checkStack(vxdHandle);
  if ((os == WNT) && params.dumpDriver) params.dumpDriverMemory(0);
  DWORD ourXid = params.fillSendBuf(sendBuf, vxdHandle);
  params.sock = params.getBootpSocket();
  ReceiverParams receiverParams(params.sock, &recvBuf, ourXid,
                                params.mainWindow,
                                vxdHandle, params.promiscuous,
                                params.rxTimeout);
  HANDLE receiverHandle =
    HANDLE(_beginthread(bootpReceiver, ThreadStackSize, &receiverParams));
  SenderParams senderParams(params.sock, &sendBuf, params.mainWindow);
  HANDLE senderHandle = HANDLE(_beginthread(bootpSender, ThreadStackSize,
                                            &senderParams));
  if (WaitForSingleObject(receiverHandle, params.rxTimeout) == WAIT_OBJECT_0) {
    quiesceSocket(params.sock, FALSE); // Shouldn't be needed to stop sender.
    CloseHandle(senderHandle);
    if (receiverParams.success) {
      if (params.registerResponse(recvBuf)) params.allowReboot();
    }
    else fail(params.mainWindow, "No valid response received.");
  }
  else {                        // Receiver thread hasn't finished.
    CloseHandle(receiverHandle);
    quiesceSocket(params.sock, FALSE);
    CloseHandle(senderHandle);
    WaitForSingleObject(receiverHandle, INFINITE); // probably invalid handle
    if (!params.promiscuous) {
      logLine(params.mainWindow, "No valid response received.");
      if (!params.disconnected()) params.handleBootpFailure(vxdHandle);
    }
  }
  WaitForSingleObject(senderHandle, INFINITE); // probably invalid handle
#if 0                           // ***temp***
  PostMessage(params.mainWindow, WM_CLOSE, 0, 0);
#else  // lets service-launched app finish
  SendMessage(params.mainWindow, WM_CLOSE, 0, 0);
#endif
  _endthread();
}

int APIENTRY mainDialogProc(HWND hDlg, UINT message, UINT wParam,
                            LONG lParam)
{
  switch (message)
    {
    case WM_CLOSE:
      writeLog(hDlg);
      EndDialog (hDlg, 0);
      PostQuitMessage(0);
      break;
    }
  return DefWindowProc (hDlg, message, wParam, lParam);
}

/*
   Here we set up the GUI.  The use of ShowWindow() is probably inept, but
   it's the only way I could figure out that would allow the nCmdShow arg
   to WinMain() to coerce this program into running minimized if the user
   wants it that way.  The problem seems to be that nCmdShow is always
   the same (SW_SHOWDEFAULT == decimal 10) no matter what the Win95 Shortcut
   Run property is.  Am I missing something or what?

   Another idisyncrasy here is that we disable the Close item on the control
   menu if we're running under NT with the NT driver.  This is because we
   want to be sure any outstanding read gets successfully canceled so the
   driver can be dismissed.  This isn't appropriate, though, in case we want
   the user to see the window and dismiss it only after he's seen enough,
   so we make it an option.
*/

BOOL initGUI(HANDLE hInstance, HWND& amainWindow, int nCmdShow,
             BOOL allowClose)
{
  static char appName[] = "BILLGPC";
  WNDCLASS wndclass;
  wndclass.style = CS_HREDRAW | CS_VREDRAW;
  wndclass.lpfnWndProc = (WNDPROC)mainDialogProc;
  wndclass.cbClsExtra = 0;
  wndclass.cbWndExtra = DLGWINDOWEXTRA;
  wndclass.hInstance = hInstance;
  wndclass.hIcon = LoadIcon(hInstance, "ICON_3");
  wndclass.hCursor = LoadCursor (NULL, IDC_ARROW);
  wndclass.hbrBackground  = (HBRUSH)(COLOR_BTNFACE+1);
  wndclass.lpszMenuName = NULL;
  wndclass.lpszClassName = appName;
  RegisterClass(&wndclass);
  amainWindow = CreateDialog(hInstance, MAKEINTRESOURCE(DIALOG_MAIN), NULL,
                             (DLGPROC)mainDialogProc);
  if (amainWindow) {
    if ((srvHandle != NULL) && !allowClose) {
      HMENU ctrlMenuH = GetSystemMenu(amainWindow, FALSE);
      if (ctrlMenuH != NULL) {
        EnableMenuItem(ctrlMenuH, SC_CLOSE, MF_BYCOMMAND | MF_GRAYED);
      }
    }
    ShowWindow(amainWindow,
               ((nCmdShow == SW_SHOWDEFAULT) ? SW_SHOWMINIMIZED : nCmdShow));
    return TRUE;
  }
  else {
    char* complaint = "Couldn't create main window";
    outFile << complaint << endl;
    MessageBox(0, complaint, "billgPC", MB_ICONSTOP);
    return FALSE;
  }
}

/*
  Here we extract the /t<seconds> option, returning a millisecond value,
  or, failing that, return the default 30,000 milliseconds.
*/

DWORD getTimeout(char* acmdLine)
{
  DWORD result = 30000;
  if (acmdLine != NULL) {
    char* optionP = strstr(acmdLine, "/t");
    if (optionP == NULL) optionP = strstr(acmdLine, "/T");
    if (optionP != NULL) {
      int secs = atoi(optionP + 2);
      if (secs > 0) result = secs * 1000;
    }
  }
  return result;
}

/*
  Here we extract the /h<hopcount> option, returning an initial value
  that might fool Novell Netware's IP stack
  or, failing that, return the default 0.
*/

DWORD getHops(char* acmdLine)
{
  DWORD result = 0;
  if (acmdLine != NULL) {
    char* optionP = strstr(acmdLine, "/h");
    if (optionP == NULL) optionP = strstr(acmdLine, "/H");
    if (optionP != NULL) {
      int hops = atoi(optionP + 2);
      if (hops > 0) result = hops;
    }
  }
  return result;
}

/*
  We want to set the current directory to the one we were launched from.
  This can be tricky because if we were launched by the billgPC Launcher
  our full file path is encased in quotes in the command line.  Then we can
  open the logfile according to whether the user wants to overwrite any old
  logfile or append to it.  If we were launched from the command line
  without an explicit path, there's no need to set the current directory.
*/

void openLog(char* cmdLine)
{
  char* lastBackslash = strchr(cmdLine, '\\');
  while (lastBackslash != NULL) {
    size_t nextSentinelLoc = strcspn(lastBackslash + 1, "\\ ");
    char sentinel = lastBackslash[nextSentinelLoc + 1];
    if (sentinel != '\\') break;
    lastBackslash += (nextSentinelLoc + 1);
  }
  char dirPath[MAX_PATH];
  if (lastBackslash != NULL) {
    char* cmdLineDirStart = cmdLine + strspn(cmdLine, "\"");
    size_t dirPathLen = lastBackslash - cmdLineDirStart;
    memcpy(dirPath, cmdLineDirStart, dirPathLen);
    dirPath[dirPathLen] = '\0';
    SetCurrentDirectory(dirPath);
  }
  if ((strstr(cmdLine, "/a") != NULL) || (strstr(cmdLine, "/A") != NULL)) {
    outFile.open("BILLGPC.LOG",ios::out | ios::app);
  }
  else outFile.open("BILLGPC.LOG", ios::out | ios::trunc);
  char dateStr[80], timeStr[80];
  GetDateFormat(LOCALE_USER_DEFAULT, 0, NULL,
                "yyyy'-'MM'-'dd", dateStr, sizeof(dateStr));
  GetTimeFormat(LOCALE_USER_DEFAULT, TIME_FORCE24HOURFORMAT, NULL,
                "HH':'mm':'ss", timeStr, sizeof(timeStr));
  outFile << "billgPC " VERSION " started " << dateStr << " " << timeStr
          << endl;
  outFile << "Command line: " << cmdLine << endl;
}

void setupRFC1084TagTranslation(char* acmdLine)
{
  for (char* delim = strchr(acmdLine, '/');
       delim != NULL;
       delim = strchr(delim + 1, '/')) {
    size_t aliasTag, realTag;
    if ((sscanf(delim, "/%u=%u", &aliasTag, &realTag) == 2) &&
        (aliasTag < sizeof(RFC1084Field::tagXlateTbl)) &&
        (realTag < sizeof(RFC1084Field::tagXlateTbl))) {
      RFC1084Field::tagXlateTbl[aliasTag] = UCHAR(realTag);
    }
  }
}

/*
   Here we marshall the options entered from the command line.  We also
   set the current directory to be the one from which the program was loaded.

   Originally we set the current directory so we could find the VxD even if
   we'd been started from the Registry key
   HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Run.
   This reason became obsolete when billgps.vxd became an NDIS protocol
   installed so that it gets loaded by the OS from the system directory,
   but we left the logic in because it made billgpc.log much easier to
   find when a remote user sent email complaining that the program didn't
   work.
*/

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
                   LPSTR /* lpCmdLine */, int nCmdShow)
{
  ourInstance = hInstance;
  HWND mainWindow = 0;          // our main window; 0 means not yet ready
  BOOL forceFix = FALSE;        // force fixing of Registry discrepancies?
  BOOL promiscuous = FALSE;     // accept the first packet no matter what?
  BOOL dumpDriver = FALSE;      // dump the driver?
  BOOL forceIEEE = FALSE;       // force IEEE htype in bootp request?
  BOOL forceEthernet = FALSE;   // force Ethernet htype in bootp request?
  BOOL forceReboot;             // force reboot on Registry change?
  BOOL askBroadcast;            // ask for broadcast reply first time?
  BOOL updateComputerName;      // Hostname coerces ComputerName
  BOOL ignoreHostName;          // Hostname gets ignored.
  char* cmdLine = GetCommandLine();
  if ((cmdLine != NULL) && (*cmdLine != '\0')) {
    openLog(cmdLine);
    setupRFC1084TagTranslation(cmdLine);
    forceFix = BOOL((strstr(cmdLine, "/f") != NULL) ||
                    (strstr(cmdLine, "/F") != NULL));
    promiscuous = BOOL((strstr(cmdLine, "/p") != NULL) ||
                       (strstr(cmdLine, "/P") != NULL));
    dumpDriver = BOOL((strstr(cmdLine, "/d") != NULL) ||
                      (strstr(cmdLine, "/D") != NULL));
    forceIEEE = BOOL((strstr(cmdLine, "/i") != NULL) ||
                     (strstr(cmdLine, "/I") != NULL));
    forceEthernet = BOOL((strstr(cmdLine, "/e") != NULL) ||
                         (strstr(cmdLine, "/E") != NULL));
    forceReboot = BOOL((strstr(cmdLine, "/r") != NULL) ||
                       (strstr(cmdLine, "/R") != NULL));
    askBroadcast = BOOL((strstr(cmdLine, "/b") != NULL) ||
                        (strstr(cmdLine, "/B") != NULL));
    updateComputerName = BOOL((strstr(cmdLine, "/n") != NULL) ||
                              (strstr(cmdLine, "/N") != NULL));
    ignoreHostName = BOOL((strstr(cmdLine, "/xh") != NULL) ||
                          (strstr(cmdLine, "/XH") != NULL));
  }
  else {
    MessageBox(0, "Can't get command line", "billgPC", MB_ICONSTOP);
    exit(-1);
  }
  configureForOS(nCmdShow); // Do this before any screen I/O!
  if ((!hPrevInstance) &&
      initGUI(hInstance, mainWindow, nCmdShow,
              BOOL(promiscuous || dumpDriver))) {
    logLine(mainWindow, "billgPC Version "  VERSION);
    logLine(mainWindow, "Free software copyright (C) 1996-8 Lew Perin");
    if (askBroadcast) logLine(mainWindow, "Asking for broadcast first time");
    WorkerParams workerParams(mainWindow, forceFix, promiscuous, dumpDriver,
                              forceIEEE, forceReboot, askBroadcast,
                              getTimeout(cmdLine),
                              forceEthernet,
                              updateComputerName,
                              getHops(cmdLine),
                              ignoreHostName);
    _beginthread(worker, ThreadStackSize, &workerParams);
    MSG msg;
    while (GetMessage(&msg, NULL, 0, 0)) {
      if (!IsDialogMessage(mainWindow, &msg)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
      }
    }
  }
  shutdown();
  return 0;
}
