// regdump.cpp

#include <windows.h>
#include <iostream.h>
#include <iomanip.h>

void usage()
{
  cerr << "usage: regdump " << "key-path" << endl;
  exit(-1);
}

// Table listing info on legal root keys:

struct { 
  char* name;
  HKEY handle;
} rk[] = {
  { "HKEY_CLASSES_ROOT", HKEY_CLASSES_ROOT },
  { "HKEY_CURRENT_USER", HKEY_CURRENT_USER },
  { "HKEY_LOCAL_MACHINE", HKEY_LOCAL_MACHINE },
  { "HKEY_USERS", HKEY_USERS },
  { NULL, NULL }
};

BOOL isRootKey(HKEY h)
{
  for (int i = 0; rk[i].name != NULL; i++) {
    if (rk[i].handle == h) return TRUE;
  }
  return FALSE;
}

HKEY getRootKey(const char* apath)
{
  for (int i = 0; rk[i].name != NULL; i++) {
    size_t rkiLen = strlen(rk[i].name);
    if (!strncmp(apath, rk[i].name, rkiLen)) {
      if ((apath[rkiLen] == '\0') || (apath[rkiLen] == '\\')) {
        return rk[i].handle;
      }
    }
  }
  return NULL;
}

HKEY getKey(const char* apath)
{
  HKEY thisKeyHandle, nextKeyHandle;
  if ((thisKeyHandle = getRootKey(apath)) == NULL) usage();
  char subKeyName[80];
  const char* separatorP = apath + strcspn(apath, "\\");
  while (*separatorP) {
    const char* nextSubKeyP = separatorP + 1;
    size_t subKeyLen = strcspn(nextSubKeyP, "\\");
    memcpy(subKeyName, nextSubKeyP, subKeyLen);
    subKeyName[subKeyLen] = '\0';
    separatorP = nextSubKeyP + subKeyLen;
    if (RegOpenKeyEx(thisKeyHandle, subKeyName, 0, KEY_READ,
                     &nextKeyHandle)!= ERROR_SUCCESS) {
      cout << endl << "*** Not found:" << endl << apath << endl
           << "***" << endl;
      exit(0);
    }
    if (!isRootKey(thisKeyHandle)) RegCloseKey(thisKeyHandle);
    thisKeyHandle = nextKeyHandle;
  }
  return thisKeyHandle;
}

void dumpOneValue(HKEY akeyHandle, DWORD avalueNo, DWORD avalueNameSize,
                  DWORD avalueSize)
{
  DWORD dataType;
  DWORD inputValueSize = avalueSize;
  CHAR* valueName = new CHAR[avalueNameSize];
  BYTE* value = new BYTE[avalueSize];
  LONG rc = RegEnumValue(akeyHandle, avalueNo, valueName, &avalueNameSize,
                         NULL, &dataType, value, &avalueSize);
  DWORD safeVSize = (avalueSize < inputValueSize) ? // Thanks, Microsoft!
    avalueSize : inputValueSize;
  if ((rc == ERROR_SUCCESS) || (rc == ERROR_MORE_DATA)) {
    cout << ((*valueName == '\0') ? "@" : valueName) << " is ";
    switch(dataType)
      {
      case REG_BINARY:
        cout << "REG_BINARY (hex): ";
        {
          for (int i = 0; i < safeVSize - 1; i++) {
            cout << hex << setw(2) << setfill('0') << WORD(value[i]) << ".";
          }
          cout << hex << setw(2) << setfill('0') << WORD(value[safeVSize - 1]);
        }
        break;
      case REG_DWORD_BIG_ENDIAN:
        cout << "REG_DWORD_BIG_ENDIAN (hex): ";
        {
          for (int i = 0; i < safeVSize - 1; i++) {
            cout << hex << setw(2) << setfill('0') << WORD(value[i]) << ".";
          }
          cout << hex << setw(2) << setfill('0') << WORD(value[safeVSize - 1]);
        }
        break;
      case REG_DWORD_LITTLE_ENDIAN:
        cout << "REG_DWORD_LITTLE_ENDIAN (hex): " << hex << *((DWORD*)value);
        break;
      case REG_EXPAND_SZ:
        cout << "REG_EXPAND_SZ: " << (CHAR*)value;
        break;
      case REG_SZ:
        cout << "REG_SZ: " << (CHAR*)value;
        break;
      case REG_LINK:
        cout << "(REG_LINK)";
        break;
      case REG_MULTI_SZ:
        cout << "REG_MULTI_SZ:";
        {
          for (CHAR* thisValue = (CHAR*)value;
               *thisValue;
               thisValue += (strlen(thisValue) + 1)) {
            cout << endl << "  " << thisValue;
          }
        }
        break;
      case REG_NONE:
        cout << "(REG_NONE)";
        break;
      case REG_RESOURCE_LIST:
        cout << "(REG_RESOURCE_LIST)";
        break;
      default:
        cout << "UNKNOWN DATA TYPE (" << dataType << ")";
        break;
      }
    if (rc == ERROR_MORE_DATA) {
      cout << " (ERROR_MORE_DATA with data size " << inputValueSize
        << " in, " << avalueSize << " out)";
    }
    cout << endl;
  }
  else cout << "(Couldn't dump value due to error " << hex << rc << ")" << endl;
  delete valueName;
  delete value;
  cout << dec;
}

void dumpKey(char* akeyName)
{
  enum { KeyClassSize = 80 };
  DWORD subKeyNameSize;
  DWORD nSubKeys, nValues, maxSubKeyNameSize, maxValueNameSize;
  DWORD maxValueDataSize, keyClassSize = KeyClassSize;
  CHAR keyClass[KeyClassSize];
  FILETIME lastUpdate;
  HKEY keyHandle = getKey(akeyName);


  if (RegQueryInfoKey(keyHandle, keyClass, &keyClassSize, NULL, &nSubKeys, 
                      &maxSubKeyNameSize, NULL, &nValues, &maxValueDataSize,
                      &maxValueNameSize, NULL, NULL) !=  ERROR_SUCCESS) {
    cerr << "Couldn't query info for key" << endl;
    exit(-1);
  }

  // Because RegQueryInfoKey sometimes underestimates max name or value size,
  // and because one value we found on an NT box was 3750 bytes:
  if (maxValueNameSize < 80) maxValueNameSize = 80;
  if (maxValueDataSize < 4095) maxValueDataSize = 4095;
  if (maxSubKeyNameSize < 80) maxSubKeyNameSize = 80;

  CHAR* subKeyName = new CHAR[maxSubKeyNameSize + 1];

  cout << endl << "Key: " << akeyName << endl;
  if (keyClassSize) cout << "Class: " << keyClass << endl;
  cout << endl;

  if (nValues) {  // show all the values
    for (DWORD valueNo = 0; valueNo < nValues; valueNo++) {
      dumpOneValue(keyHandle, valueNo, maxValueNameSize + 1,
                   maxValueDataSize + 1);
    }
  }

  for (DWORD n = 0; ; n++) { // Show all the subkeys - don't trust nSubKeys
    subKeyNameSize = maxSubKeyNameSize;
    LONG rc = RegEnumKeyEx(keyHandle, n, subKeyName,
                           &subKeyNameSize, NULL, NULL, NULL, &lastUpdate);
    if (rc != ERROR_SUCCESS) {
      if (rc != ERROR_NO_MORE_ITEMS) {
        cerr << "Couldn't query info for subkey " << n << ": " << rc <<  endl;
      }
      break;
    }
    char* keyName = new char[strlen(akeyName) + strlen(subKeyName) + 3];
    strcpy(keyName, akeyName);
    strcat(keyName, "\\");
    strcat(keyName, subKeyName);
    dumpKey(keyName);
    delete keyName;
  }

  // close the key
  if (!isRootKey(keyHandle)) {
    if (RegCloseKey(keyHandle) != ERROR_SUCCESS) {
      cerr << "Unable to close subkey" << endl;
      exit(-1);
    }
  }

  delete subKeyName;
}

/*
  We use Windows's command-line-in-one-gulp function because a Registry
  key may contain embedded whitespace.  We just want everything after
  the name of the executable.
*/
int main(int /* argc */, char** /*argv */)
{
  LPTSTR commandLine = GetCommandLine();
  if (commandLine != NULL) {
    char* clp = strchr(commandLine, ' '); // up over exe name
    if (clp != NULL) {
      while (*clp == ' ') clp++; // ...and any spaces.
      char* clpTrailer = clp + strlen(clp) - 1;
      while ((clpTrailer > clp) && (*clpTrailer <= ' ')) {
        *clpTrailer-- = '\0';   // Trim trailing whitespace.
      }
      dumpKey(clp);
    }
    else usage();
  }
  else usage();

  return 0;
}
